######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Exporter's specifications."""
from dataclasses import dataclass
from enum import Enum, unique
from spine_engine.project_item.project_item_specification import ProjectItemSpecification
from spinedb_api.export_mapping.export_mapping import (
    DefaultValueIndexNameMapping,
    EntityClassMapping,
    EntityGroupEntityMapping,
    EntityGroupMapping,
    ExportMapping,
    FixedValueMapping,
    IndexNameMapping,
    ParameterDefaultValueIndexMapping,
    ParameterValueIndexMapping,
)
from spinedb_api.export_mapping.export_mapping import from_dict as mapping_from_dict
from spinedb_api.export_mapping.export_mapping import legacy_group_fn_from_dict
from spinedb_api.mapping import Position
from spinedb_api.mapping import to_dict as mapping_to_dict
from spinedb_api.mapping import unflatten
from .item_info import ItemInfo


@unique
class MappingType(Enum):
    alternatives = "alternatives"
    entities = "entities"
    entity_groups = "entity_groups"
    entity_parameter_default_values = "entity_parameter_default_values"
    entity_parameter_values = "entity_parameter_values"
    entity_dimension_parameter_default_values = "entity_dimension_parameter_default_values"
    entity_dimension_parameter_values = "entity_dimension_parameter_values"
    parameter_value_lists = "parameter_value_lists"
    scenario_alternatives = "scenario_alternatives"
    scenarios = "scenarios"

    @classmethod
    def from_legacy_type(cls, type_str):
        type_str = {
            "objects": "entities",
            "object_groups": "entity_groups",
            "object_default_parameter_values": "entity_parameter_default_values",
            "object_parameter_values": "entity_parameter_values",
            "relationships": "entities",
            "relationship_default_parameter_values": "entity_parameter_default_values",
            "relationship_parameter_values": "entity_parameter_values",
            "relationship_object_parameter_default_values": "entity_dimension_parameter_default_values",
            "relationship_object_parameter_values": "entity_dimension_parameter_values",
        }.get(type_str, type_str)
        return cls(type_str)


@unique
class OutputFormat(Enum):
    CSV = "csv"
    EXCEL = "Excel"
    GDX = "gdx"
    SQL = "SQL"

    def is_compatible_file_extension(self, extension):
        """Tests if given file extension is acceptable for current output format.

        Args:
            extension (str): file extension without the dot
        """
        if self == OutputFormat.CSV:
            return extension in ("csv", "dat", "txt")
        if self == OutputFormat.EXCEL:
            return extension == "xlsx"
        if self == OutputFormat.GDX:
            return extension == "gdx"
        if self == OutputFormat.SQL:
            return extension in ("sqlite", "sqlite3")
        return False

    def file_extension(self):
        """Returns a file extension for the output format.

        Returns:
            str: file extension without the dot
        """
        return {
            OutputFormat.CSV: "csv",
            OutputFormat.EXCEL: "xlsx",
            OutputFormat.GDX: "gdx",
            OutputFormat.SQL: "sqlite",
        }[self]

    @staticmethod
    def output_format_from_extension(extension):
        """Creates an output format from given file extension.

        Args:
            extension (str): file extension without the dot

        Returns:
            OutputFormat: output format or None if extension is unknown
        """
        try:
            return {
                "csv": OutputFormat.CSV,
                "dat": OutputFormat.CSV,
                "gdx": OutputFormat.GDX,
                "sqlite": OutputFormat.SQL,
                "txt": OutputFormat.CSV,
                "xlsx": OutputFormat.EXCEL,
            }[extension]
        except KeyError:
            return None

    @staticmethod
    def default():
        """Return the default output format.

        Returns:
            OutputFormat: default output format
        """
        return OutputFormat.CSV

    def is_multi_file_capable(self):
        """Tests if the output format is capable of having multiple output files.

        Returns:
            bool: True if multiple output files are possible, False otherwise
        """
        return self == OutputFormat.CSV


@dataclass(eq=False)
class MappingSpecification:
    type: MappingType
    enabled: bool
    always_export_header: bool
    group_fn: str
    use_fixed_table_name_flag: bool
    root: ExportMapping

    def to_dict(self):
        """Serializes specification into dictionary.

        Returns:
            dict: serialized specification
        """
        mapping_dict = dict(vars(self))
        mapping_dict["type"] = mapping_dict["type"].value
        mapping_dict["root"] = mapping_to_dict(mapping_dict["root"])
        return mapping_dict

    @staticmethod
    def from_dict(mapping_dict):
        """Deserializes specification from dictionary.

        Args:
            mapping_dict (dict): serialized specification

        Returns:
            MappingSpecification: deserialized specification
        """
        root = mapping_from_dict(mapping_dict.pop("root"))
        type_ = MappingType.from_legacy_type(mapping_dict.pop("type"))
        return MappingSpecification(root=root, type=type_, **mapping_dict)


class Specification(ProjectItemSpecification):
    """Exporter's specification."""

    def __init__(self, name="", description="", mapping_specifications=None, output_format=OutputFormat.default()):
        """
        Args:
            name (str): specification name
            description (str): description
            mapping_specifications (dict, optional): mapping from export mapping name to ``MappingSpecification``
            output_format (OutputFormat): output format
        """
        super().__init__(name, description, ItemInfo.item_type())
        if mapping_specifications is None:
            mapping_specifications = {}
        self._mapping_specifications = mapping_specifications
        self.output_format = output_format

    def is_equivalent(self, other):
        """
        Checks if specifications are essentially equivalent.

        Args:
            other (Specification): another specification

        Returns:
            bool: True if specifications are equivalent, False otherwise
        """
        return (
            self.output_format == other.output_format and self._mapping_specifications == other._mapping_specifications
        )

    def mapping_specifications(self):
        """
        Returns export mapping specifications.

        Returns:
            dict: mapping specifications
        """
        return self._mapping_specifications

    def enabled_specifications(self):
        """
        Returns enabled export mapping specifications.

        Returns:
            dict: mapping specifications
        """
        return {name: spec for name, spec in self._mapping_specifications.items() if spec.enabled}

    def root_mapping(self, name):
        """
        Returns root mapping for given name.

        Args:
            name (str): mapping's name

        Returns:
            Mapping: root mapping
        """
        return self._mapping_specifications[name].root

    def mapping_type(self, name):
        """
        Returns mapping type for given name.

        Args:
            name (str): mapping's name

        Returns:
            MappingType: mapping's type
        """
        return self._mapping_specifications[name].type

    def is_exporting_multiple_files(self):
        """Tests if this specification would result in multiple files being exported.

        Returns:
            bool: True if multiple files will be exporter, False otherwise
        """
        if not self.output_format.is_multi_file_capable():
            return False
        for mapping_spec in self._mapping_specifications.values():
            if mapping_spec.use_fixed_table_name_flag or any(
                m.position == Position.table_name for m in mapping_spec.root.flatten()
            ):
                return True
        return False

    def to_dict(self):
        """
        Serializes the specification into JSON compatible dictionary.

        Returns:
            dict: serialized specification
        """
        mappings = {}
        for name, mapping_spec in self._mapping_specifications.items():
            spec_dict = {
                "type": mapping_spec.type.value,
                "mapping": mapping_to_dict(mapping_spec.root),
                "enabled": mapping_spec.enabled,
                "always_export_header": mapping_spec.always_export_header,
                "group_fn": mapping_spec.group_fn,
                "use_fixed_table_name": mapping_spec.use_fixed_table_name_flag,
            }
            mappings[name] = spec_dict
        return {
            "item_type": ItemInfo.item_type(),
            "output_format": self.output_format.value,
            "name": self.name,
            "description": self.description,
            "mappings": mappings,
        }

    @staticmethod
    def from_dict(specification_dict):
        """
        Restores specification from dictionary.

        Args:
            specification_dict (dict): serialized specification

        Returns:
            Specification: restored specification
        """
        for spec_dict in specification_dict["mappings"].values():
            # Legacy: remove parameter value mappings from object group mappings,
            # they're not supported anymore.
            if spec_dict["type"] == "object_group_parameter_values":
                spec_dict["type"] = MappingType.entity_groups.value
                keep_map_types = {
                    FixedValueMapping.MAP_TYPE,
                    EntityClassMapping.MAP_TYPE,
                    EntityGroupMapping.MAP_TYPE,
                    EntityGroupEntityMapping.MAP_TYPE,
                }
                spec_dict["mapping"] = [m for m in spec_dict["mapping"] if m["map_type"] in keep_map_types]
        mapping_specifications = {
            name: MappingSpecification(
                MappingType.from_legacy_type(spec_dict["type"]),
                spec_dict.get("enabled", True),
                spec_dict.get("always_export_header", True),
                spec_dict.get("group_fn", legacy_group_fn_from_dict(spec_dict["mapping"])),
                spec_dict.get("use_fixed_table_name", False),
                _add_index_names(mapping_from_dict(spec_dict["mapping"])),
            )
            for name, spec_dict in specification_dict["mappings"].items()
        }
        try:
            output_format = OutputFormat(specification_dict["output_format"])
        except ValueError:
            output_format = OutputFormat.default()
        return Specification(
            specification_dict["name"], specification_dict["description"], mapping_specifications, output_format
        )


def _add_index_names(mapping):
    """Adds index name mappings to legacy mappings that don't have them.

    Args:
        mapping (ExportMapping): mapping to modify

    Returns:
        mapping (ExportMapping): fixed mapping
    """
    flattened = mapping.flatten()
    if any((isinstance(m, (IndexNameMapping, DefaultValueIndexNameMapping)) for m in flattened)):
        return mapping
    if any((isinstance(m, (ParameterValueIndexMapping, ParameterDefaultValueIndexMapping)) for m in flattened)):
        fixed = []
        for m in flattened:
            if isinstance(m, ParameterValueIndexMapping):
                fixed.append(IndexNameMapping(Position.hidden))
            elif isinstance(m, ParameterDefaultValueIndexMapping):
                fixed.append(DefaultValueIndexNameMapping(Position.hidden))
            fixed.append(m)
        mapping = unflatten(fixed)
    return mapping



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for the OptionsWidget class."""
import unittest
from unittest.mock import MagicMock
from PySide6.QtGui import QUndoStack
from PySide6.QtWidgets import QApplication
from spine_items.importer.widgets.options_widget import OptionsWidget


class TestOptionsWidget(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._undo_stack = QUndoStack()

    def test_spin_box_change_signalling(self):
        option_template = {"number": {"label": "int value", "type": int, "default": 0}}
        connector = MagicMock()
        connector.connection.BASE_OPTIONS = {}
        connector.connection.OPTIONS = option_template
        widget = OptionsWidget(self._undo_stack)
        widget.set_connector(connector)
        changed_options = {}
        widget.options_changed.connect(changed_options.update)
        spin_box = widget.cellWidget(0, 0).layout().itemAt(1).widget()
        spin_box.setValue(23)
        connector.update_options.assert_called_once_with({"number": 23})
        self.assertEqual(changed_options, {"number": 23})

    def test_line_edit_change_signalling(self):
        option_template = {"text": {"label": "text value", "type": str, "default": ""}}
        connector = MagicMock()
        connector.connection.BASE_OPTIONS = {}
        connector.connection.OPTIONS = option_template
        widget = OptionsWidget(self._undo_stack)
        widget.set_connector(connector)
        changed_options = {}
        widget.options_changed.connect(changed_options.update)
        line_edit = widget.cellWidget(0, 0).layout().itemAt(1).widget()
        line_edit.setText("the text has been set")
        connector.update_options.assert_called_once_with({"text": "the text has been set"})
        self.assertEqual(changed_options, {"text": "the text has been set"})

    def test_combo_box_change_signalling(self):
        option_template = {
            "choice": {"label": "a choice", "type": list, "Items": ["choice a", "choice b"], "default": "choice a"}
        }
        connector = MagicMock()
        connector.connection.BASE_OPTIONS = {}
        connector.connection.OPTIONS = option_template
        widget = OptionsWidget(self._undo_stack)
        widget.set_connector(connector)
        changed_options = {}
        widget.options_changed.connect(changed_options.update)
        combo_box = widget.cellWidget(0, 0).layout().itemAt(1).widget()
        combo_box.setCurrentText("choice b")
        connector.update_options.assert_called_once_with({"choice": "choice b"})
        self.assertEqual(changed_options, {"choice": "choice b"})

    def test_check_box_change_signalling(self):
        option_template = {"yesno": {"label": "check me", "type": bool, "default": True}}
        connector = MagicMock()
        connector.connection.BASE_OPTIONS = {}
        connector.connection.OPTIONS = option_template
        widget = OptionsWidget(self._undo_stack)
        widget.set_connector(connector)
        changed_options = {}
        widget.options_changed.connect(changed_options.update)
        check_box = widget.cellWidget(0, 0).layout().itemAt(1).widget()
        check_box.setChecked(False)
        connector.update_options.assert_called_once_with({"yesno": False})
        self.assertEqual(changed_options, {"yesno": False})

    def test_set_connector_makes_copy_of_connections_base_options(self):
        option_template = {"yesno": {"label": "check me", "type": bool, "default": True}}
        connector = MagicMock()
        connector.connection.BASE_OPTIONS = {"text": {"label": "text value", "type": str, "default": ""}}
        connector.connection.OPTIONS = option_template
        widget = OptionsWidget(self._undo_stack)
        widget.set_connector(connector)
        self.assertEqual(
            connector.connection.BASE_OPTIONS, {"text": {"label": "text value", "type": str, "default": ""}}
        )


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`ParameterTreeWidget`."""
import pickle
from PySide6.QtCore import QMimeData
from PySide6.QtWidgets import QMessageBox, QTreeWidget, QTreeWidgetItem
from spinedb_api import DatabaseMapping, SpineDBAPIError
from .drop_target_table import DROP_MIME_TYPE


class ParameterTreeWidget(QTreeWidget):
    """A tree widget with drag capabilities to show entity classes and parameters."""

    def mimeData(self, items):
        mime_data = QMimeData()
        parameters = {}
        for item in items:
            if item.parent() is None:
                for i in range(item.childCount()):
                    parameters.setdefault(item.text(0), []).append(item.child(i).text(0))
            else:
                parameters.setdefault(item.parent().text(0), []).append(item.text(0))
        mime_data.setData(DROP_MIME_TYPE, pickle.dumps(parameters))
        return mime_data

    def load_data(self, url):
        """
        Loads parameter data from given URL.

        Args:
            url (str): database URL
        """
        parameters = {}
        try:
            with DatabaseMapping(url) as db_map:
                try:
                    for definition_row in db_map.query(db_map.entity_parameter_definition_sq):
                        parameters.setdefault(definition_row.entity_class_name, []).append(
                            definition_row.parameter_name
                        )
                except SpineDBAPIError as error:
                    QMessageBox.information(
                        self, "Error while reading database", f"Could not read from database {url}:\n{error}"
                    )
        except SpineDBAPIError as error:
            QMessageBox.information(self, "Error while opening database", f"Could not open database {url}:\n{error}")
            return
        self.clear()
        for class_name, parameter_names in parameters.items():
            class_item = QTreeWidgetItem([class_name])
            for name in parameter_names:
                QTreeWidgetItem(class_item, [name])
            self.addTopLevelItem(class_item)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Tool's specification factory."""
from spine_engine.project_item.project_item_specification_factory import ProjectItemSpecificationFactory
from .item_info import ItemInfo
from .tool_specifications import make_specification


class SpecificationFactory(ProjectItemSpecificationFactory):
    """A factory to make tool specifications."""

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    @staticmethod
    def make_specification(definition, app_settings, logger):
        """Returns a tool specifications."""
        return make_specification(definition, app_settings, logger)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`ParameterRenamesTableModel`."""
from enum import IntEnum, unique
from PySide6.QtCore import QModelIndex, Qt
from ..commands import SetData
from .parameter_drop_target_table_model import ParameterDropTargetTableModel


@unique
class RenamesTableColumn(IntEnum):
    CLASS = 0
    PARAMETER = 1
    NEW_NAME = 2


@unique
class RenamesRoles(IntEnum):
    SILENT_EDIT = Qt.ItemDataRole.UserRole + 1


class ParameterRenamesTableModel(ParameterDropTargetTableModel):
    GET_DATA_ROLES = (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.DisplayRole)
    SET_DATA_ROLES = (RenamesRoles.SILENT_EDIT, RenamesRoles.SILENT_EDIT, RenamesRoles.SILENT_EDIT)

    def __init__(self, settings, undo_stack, parent):
        """
        Args:
            settings (ParameterRenamingSettings, optional): initial settings
            undo_stack (QUndoStack): undo stack
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._undo_stack = undo_stack
        if settings is not None:
            self._renames = [
                [klass, param, new_name]
                for klass, param_renames in settings.name_map.items()
                for param, new_name in param_renames.items()
            ]
        else:
            self._renames = []

    def columnCount(self, parent=QModelIndex()):
        return 3

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        return self._renames[index.row()][index.column()]

    def flags(self, index):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDropEnabled | Qt.ItemIsEditable

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return ("Class", "Parameter", "New name")[section]
        return None

    def insertRows(self, row, count, parent=QModelIndex()):
        rows = count * [["", "", ""]]
        self.beginInsertRows(parent, row, row + count - 1)
        self._renames = self._renames[:row] + rows + self._renames[row:]
        self.endInsertRows()
        return True

    @staticmethod
    def _make_drop_row(entity_class, parameter):
        return [entity_class, parameter, ""]

    def removeRows(self, row, count, parent=QModelIndex()):
        self.beginRemoveRows(parent, row, row + count - 1)
        self._renames = self._renames[:row] + self._renames[row + count :]
        self.endRemoveRows()
        return True

    def rowCount(self, parent=QModelIndex()):
        return len(self._renames)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if role == Qt.ItemDataRole.EditRole:
            column = index.column()
            if column == RenamesTableColumn.CLASS:
                message = "change class name"
            elif column == RenamesTableColumn.PARAMETER:
                message = "change parameter name"
            else:
                message = "change parameter's new name"
            previous = self._renames[index.row()][column]
            if value == previous:
                return False
            self._undo_stack.push(SetData(message, index, value, previous, RenamesRoles.SILENT_EDIT))
            return True
        if role == RenamesRoles.SILENT_EDIT:
            row = index.row()
            column = index.column()
            self._renames[row][column] = value
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
            return True
        return False



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Tool project item."""
import os
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtGui import QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QApplication, QMenu
from spine_engine.config import TOOL_OUTPUT_DIR
from spine_engine.project_item.project_item_resource import ProjectItemResource
from spine_items.tool.item_info import ItemInfo
from spine_items.tool.tool import Tool
from spine_items.tool.tool_factory import ToolFactory
from spine_items.tool.tool_specifications import ExecutableTool
from tests.mock_helpers import create_mock_project, create_mock_toolbox, mock_finish_project_item_construction


class TestTool(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self._temp_dir = TemporaryDirectory()
        self.toolbox = create_mock_toolbox()
        self.project = create_mock_project(self._temp_dir.name)
        self._properties_widget = None
        specifications = [
            ExecutableTool(
                name="simple_exec",
                tooltype="executable",
                path=self._temp_dir.name,
                includes=["main.sh"],
                settings=self.toolbox.qsettings(),
                logger=self.toolbox,
                description="A simple executable tool.",
                inputfiles=["input1.csv", "input2.csv"],
                inputfiles_opt=["opt_input.csv"],
                outputfiles=["output1.csv", "output2.csv"],
                cmdline_args="<args>",
            ),
            ExecutableTool(
                name="complex_exec",
                tooltype="executable",
                path=self._temp_dir.name,
                includes=["MakeFile", "src/a.c", "src/a.h", "src/subunit/x.c", "src/subunit/x.h"],
                settings=self.toolbox.qsettings(),
                logger=self.toolbox,
                description="A more complex executable tool.",
                inputfiles=["input1.csv", "input/input2.csv"],
                inputfiles_opt=["opt/*.ini", "?abc.txt"],
                outputfiles=["output1.csv", "output/output2.csv"],
                cmdline_args="subunit",
            ),
        ]
        self.specification_dict = {x.name: x for x in specifications}
        self.project.get_specification = self.specification_dict.get
        self.toolbox.project.return_value = self.project
        self.model = _MockToolSpecModel(self.specification_dict)

    def tearDown(self):
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_item_type(self):
        self.assertEqual(Tool.item_type(), ItemInfo.item_type())

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        tool = self._add_tool()
        d = tool.item_dict()
        a = ["type", "description", "x", "y", "specification", "execute_in_work", "cmd_line_args"]
        for k in a:
            self.assertTrue(k in d, f"Key '{k}' not in dict {d}")

    def test_notify_destination(self):
        source_item = mock.NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = mock.MagicMock(return_value="Data Connection")
        tool = self._add_tool()
        tool.logger.msg = mock.MagicMock()
        tool.logger.msg_warning = mock.MagicMock()
        tool.notify_destination(source_item)
        tool.logger.msg.emit.assert_called_with(
            "Link established. Tool <b>T</b> will look for input "
            "files from <b>source name</b>'s references and data directory."
        )
        source_item.item_type = mock.MagicMock(return_value="Importer")
        tool.notify_destination(source_item)
        tool.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Importer</b> and a <b>Tool</b> has not been implemented yet."
        )
        source_item.item_type = mock.MagicMock(return_value="Data Store")
        tool.notify_destination(source_item)
        tool.logger.msg.emit.assert_called_with(
            "Link established. Data Store <b>source name</b> url will be passed to Tool <b>T</b> when executing."
        )
        source_item.item_type = mock.MagicMock(return_value="Exporter")
        tool.notify_destination(source_item)
        tool.logger.msg.emit.assert_called_with(
            "Link established. The file exported by <b>source name</b> are now available in <b>T</b>."
        )
        source_item.item_type = mock.MagicMock(return_value="Tool")
        tool.notify_destination(source_item)
        tool.logger.msg.emit.assert_called_with("Link established")
        source_item.item_type = mock.MagicMock(return_value="View")
        tool.notify_destination(source_item)
        tool.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>View</b> and a <b>Tool</b> has not been implemented yet."
        )

    def test_rename(self):
        """Tests renaming a self.tool."""
        tool = self._add_tool()
        tool.activate()
        expected_name = "ABC"
        expected_short_name = "abc"
        tool.rename(expected_name, "")
        # Check name
        self.assertEqual(expected_name, tool.name)  # item name
        self.assertEqual(expected_name, tool.get_icon().name())  # name item on Design View
        # Check data_dir
        expected_data_dir = os.path.join(self.project.items_dir, expected_short_name)
        self.assertEqual(expected_data_dir, tool.data_dir)  # Check data dir
        # Check that output_dir has been updated
        expected_output_dir = os.path.join(tool.data_dir, TOOL_OUTPUT_DIR)
        self.assertEqual(expected_output_dir, tool._output_dir)

    def test_load_tool_specification(self):
        """Test that specification is loaded into selections on Tool creation,
        and then shown in the ui when Tool is activated.
        """
        options = {"output_directory": {"type": "path", "relative": False, "path": "/somedir/outputs"}}
        item_dict = {
            "type": "Tool",
            "description": "",
            "x": 0,
            "y": 0,
            "specification": "simple_exec",
            "execute_in_work": False,
            "options": options,
        }
        tool = self._add_tool(item_dict)
        with mock.patch("spine_items.tool.tool.ToolSpecificationMenu") as mock_tool_spec_menu:
            mock_tool_spec_menu.side_effect = lambda *args: QMenu()
            tool.activate()
            self._assert_is_simple_exec_tool(tool, "/somedir/outputs".replace("/", os.sep))

    def test_save_and_restore_selections(self):
        """Test that selections are saved and restored when deactivating a Tool and activating it again."""
        options = {"output_directory": {"type": "path", "relative": True, "path": ".spinetoolbox/outputs"}}
        item_dict = {"type": "Tool", "description": "", "x": 0, "y": 0, "specification": "", "options": options}
        tool = self._add_tool(item_dict)
        expected_result_dir = os.path.abspath(os.path.join(tool._project.project_dir, ".spinetoolbox", "outputs"))
        with mock.patch("spine_items.tool.tool.ToolSpecificationMenu") as mock_tool_spec_menu:
            mock_tool_spec_menu.side_effect = lambda *args: QMenu()
            tool.activate()
            self._assert_is_no_tool(tool)
            # Set the simple_exec tool specification manually
            tool._properties_ui.comboBox_tool.textActivated.emit("simple_exec")
            tool._properties_ui.radioButton_execute_in_source.setChecked(True)
            self._assert_is_simple_exec_tool(tool, expected_result_dir)
            tool.deactivate()
            tool.activate()
            self._assert_is_simple_exec_tool(tool, expected_result_dir)

    @unittest.skip("Test doesn't work on Python 3.12 on Windows")
    def test_find_input_files(self):
        """Test that input files are correctly found in resources
        when required input files and available resources are updated"""
        separator = os.sep
        item_dict = {"type": "Tool", "description": "", "x": 0, "y": 0, "specification": "simple_exec"}
        tool = self._add_tool(item_dict)
        url1 = os.path.join(self._temp_dir.name, "more_files", "input1.csv")
        url2 = os.path.join(self._temp_dir.name, "more_files", "data.csv")
        url3 = os.path.join(self._temp_dir.name, "more filess", "input1.csv")
        url4 = os.path.join(self._temp_dir.name, "other.csv")
        url5 = os.path.join(self._temp_dir.name, "other", "input2.csv")
        url6 = os.path.join(self._temp_dir.name, "input3.csv")
        expected_urls = {"url1": url1, "url2": url2, "url3": url3, "url4": url4, "url5": url5, "url6": url6}
        for key, url in expected_urls.items():
            i = url.find(separator)
            expected_urls[key] = url[:i] + separator + url[i:]
        resources = [
            ProjectItemResource("Exporter", "file", "first", url="file:///" + url1, metadata={}, filterable=False),
            ProjectItemResource("Exporter", "url", "second", url="file:///" + url2, metadata={}, filterable=False),
            ProjectItemResource("Exporter", "file", "third", url="file:///" + url3, metadata={}, filterable=False),
            ProjectItemResource("Exporter", "url", "fourth", url="file:///" + url4, metadata={}, filterable=False),
        ]
        result = tool._find_input_files(resources)
        expected = {"input1.csv": [expected_urls["url1"], expected_urls["url3"]], "input2.csv": None}
        self.assertEqual(2, len(result))
        self.assertEqual(expected["input2.csv"], result["input2.csv"])
        self.assertTrue(expected_urls["url3"] in result["input1.csv"] or expected_urls["url1"] in result["input1.csv"])
        resources.pop(0)
        resources.append(
            ProjectItemResource("Exporter", "file", "fifth", url="file:///" + url5, metadata={}, filterable=False)
        )
        result = tool._find_input_files(resources)
        expected = {"input2.csv": [expected_urls["url5"]], "input1.csv": [expected_urls["url3"]]}
        self.assertEqual(expected, result)
        resources.append(
            ProjectItemResource("Exporter", "file", "sixth", url="file:///" + url6, metadata={}, filterable=False)
        )
        tool.specification().inputfiles = set(["input2.csv", os.path.join(self._temp_dir.name, "input3.csv")])
        result = tool._find_input_files(resources)
        expected = {
            os.path.join(self._temp_dir.name, "input3.csv"): [expected_urls["url6"]],
            "input2.csv": [expected_urls["url5"]],
        }
        self.assertEqual(expected, result)

    def _add_tool(self, item_dict=None):
        if item_dict is None:
            item_dict = {"type": "Tool", "description": "", "x": 0, "y": 0}
        factory = ToolFactory()
        tool = factory.make_item("T", item_dict, self.toolbox, self.project)
        self._properties_widget = mock_finish_project_item_construction(factory, tool, self.toolbox)
        # Set model for tool combo box
        tool._properties_ui.comboBox_tool.setModel(self.model)
        self.project.get_item.return_value = tool
        return tool

    def _assert_is_simple_exec_tool(self, tool, expected_result_dir=""):
        """Assert that the given tool has the simple_exec specification."""
        # Check tool spec
        source_files = tool.specification().includes
        input_files = tool.specification().inputfiles
        opt_input_files = tool.specification().inputfiles_opt
        output_files = tool.specification().outputfiles
        self.assertEqual(source_files, ["main.sh"])
        self.assertIn("input1.csv", input_files)
        self.assertIn("input2.csv", input_files)
        self.assertEqual(opt_input_files, {"opt_input.csv"})
        self.assertIn("output1.csv", output_files)
        self.assertIn("output2.csv", output_files)
        # Check cmdline model
        spec_args_root = tool._cmdline_args_model.invisibleRootItem().child(0)
        self.assertEqual(spec_args_root.rowCount(), 1)
        cmdline_args = spec_args_root.child(0).text()
        self.assertEqual(cmdline_args, "<args>")
        # Check ui
        combox_text = tool._properties_ui.comboBox_tool.currentText()
        in_work = tool._properties_ui.radioButton_execute_in_work.isChecked()
        in_source = tool._properties_ui.radioButton_execute_in_source.isChecked()
        result_dir = tool._properties_ui.lineEdit_result_directory.text()
        self.assertEqual(expected_result_dir, result_dir)
        self.assertEqual(combox_text, "simple_exec")
        self.assertFalse(in_work)
        self.assertTrue(in_source)

    def _assert_is_no_tool(self, tool):
        """Assert that the given tool has no tool specification."""
        # Check cmdline model
        tool_args_root = tool._cmdline_args_model.invisibleRootItem().child(1)
        self.assertEqual(tool_args_root.rowCount(), 1)  # The "Type new arg here..." item
        # Check ui
        combox_text = tool._properties_ui.comboBox_tool.currentText()
        in_work = tool._properties_ui.radioButton_execute_in_work.isChecked()
        in_source = tool._properties_ui.radioButton_execute_in_source.isChecked()
        self.assertEqual(combox_text, "")
        self.assertTrue(in_work)
        self.assertFalse(in_source)


class _MockToolSpecModel(QStandardItemModel):
    # Create a dictionary of tool specifications to 'populate' the mock model
    def __init__(self, specifications):
        super().__init__()
        self._specifications = specifications
        self.invisibleRootItem().appendRows([QStandardItem(x) for x in self._specifications])


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Importer's execute kernel (do_work), as target for a multiprocess.Process"""
import os
from spine_engine.project_item.project_item_resource import get_source, get_source_extras
from spine_engine.utils.helpers import create_log_file_timestamp
from spinedb_api import InvalidMapping, clear_filter_configs
from spinedb_api.exception import ReaderError
from spinedb_api.helpers import remove_credentials_from_url
from spinedb_api.import_mapping.import_mapping_compat import parse_named_mapping_spec, unparse_named_mapping_spec
from spinedb_api.import_mapping.type_conversion import value_to_convert_spec
from spinedb_api.parameter_value import to_database
from spinedb_api.spine_db_client import SpineDBClient


def do_work(
    process, mapping, cancel_on_error, on_conflict, logs_dir, source_resources, reader, to_server_urls, lock, logger
):
    all_data = []
    all_errors = []
    table_mappings = {
        name: mappings
        for name, mappings in mapping.get("table_mappings", {}).items()
        if name in mapping["selected_tables"]
    }
    table_options = {
        name: options
        for name, options in mapping.get("table_options", {}).items()
        if name in mapping["selected_tables"]
    }
    table_column_convert_specs = {
        tn: {int(col): value_to_convert_spec(spec) for col, spec in cols.items()}
        for tn, cols in mapping.get("table_types", {}).items()
    }
    table_default_column_convert_fns = {
        tn: value_to_convert_spec(spec) for tn, spec in mapping.get("table_default_column_type", {}).items()
    }
    table_row_convert_specs = {
        tn: {int(col): value_to_convert_spec(spec) for col, spec in cols.items()}
        for tn, cols in mapping.get("table_row_types", {}).items()
    }
    to_clients = [SpineDBClient.from_server_url(server_url) for server_url in to_server_urls]
    for resource in source_resources:
        src = get_source(resource)
        if resource.hasfilepath:
            source_anchor = f"<a style='color:#BB99FF;' title='{src}' href='file:///{src}'>{os.path.basename(src)}</a>"
        else:
            safe_url = remove_credentials_from_url(src)
            source_anchor = f"<p style='color:#BB99FF;'>{safe_url}</p>"
        logger.msg.emit("Importing " + source_anchor)
        extras = get_source_extras(resource)
        try:
            reader.connect_to_source(src, **extras)
        except Exception as error:  # pylint: disable=broad-except
            logger.msg_error.emit(f"Failed to connect to {source_anchor}: {error}")
            return (False,)
        parsed_table_mappings = _parse_mappings(table_mappings)
        try:
            parsed_table_mappings = reader.resolve_values_for_fixed_position_mappings(
                parsed_table_mappings, table_options, cancel_on_error
            )
        except ReaderError as error:
            logger.msg_error.emit(f"Failed to read fixed position data in {source_anchor}: {error}")
            return (False,)
        table_mappings_with_fixed_positions = {
            table_name: [
                unparse_named_mapping_spec(mapping_name, root_mapping) for mapping_name, root_mapping in mappings
            ]
            for table_name, mappings in parsed_table_mappings.items()
        }
        for name, mappings in table_mappings_with_fixed_positions.items():
            logger.msg.emit(f"Processing table <b>{name}</b>")
            for spec in mappings:
                mapping_name = next(iter(spec.keys()))
                logger.msg.emit(f"* Applying mapping <b>{mapping_name}</b>...")
                try:
                    data, errors = reader.get_mapped_data(
                        {name: [spec]},
                        table_options,
                        table_column_convert_specs,
                        table_default_column_convert_fns,
                        table_row_convert_specs,
                        unparse_value=to_database,
                    )
                except InvalidMapping as error:
                    logger.msg_error.emit(f"Failed to import: {error}")
                    if cancel_on_error:
                        logger.msg_error.emit("Cancel import on error has been set. Bailing out.")
                        return (False,)
                    logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
                    continue
                if not errors:
                    logger.msg.emit(f"Successful ({sum(len(d) for d in data.values())} data to be written).")
                else:
                    logger.msg_warning.emit(
                        f"Read {sum(len(d) for d in data.values())} data with {len(errors)} errors."
                    )
                all_data.append(data)
                all_errors.extend(errors)
        reader.disconnect()
        if all_data:
            for client in to_clients:
                lock.acquire()
                try:
                    with process.maybe_idle:
                        client.db_checkin()
                    success = _import_data_to_url(cancel_on_error, on_conflict, logs_dir, all_data, client, logger)
                    client.db_checkout()
                    if not success and cancel_on_error:
                        return (False,)
                finally:
                    lock.release()
            all_data.clear()
    if all_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_read_error.log"))
        with open(logfilepath, "w") as f:
            for err in all_errors:
                f.write(f"{err}\n")
        # Make error log file anchor with path as tooltip
        logfile_anchor = (
            "<a style='color:#BB99FF;' title='" + logfilepath + "' href='file:///" + logfilepath + "'>Error log</a>"
        )
        logger.msg_error.emit(logfile_anchor)
        if cancel_on_error:
            logger.msg_error.emit("Cancel import on error has been set. Bailing out.")
            return (False,)
        logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
    return (True,)


def _import_data_to_url(cancel_on_error, on_conflict, logs_dir, all_data, client, logger):
    all_import_errors = []
    all_import_count = 0
    for data in all_data:
        response = client.import_data({**data, "on_conflict": on_conflict}, "")
        if "error" in response:
            all_import_errors.append(response["error"])
            continue
        import_count, import_errors = response["result"]
        all_import_count += import_count
        all_import_errors += import_errors
        if import_errors:
            logger.msg_error.emit("Errors while importing a table.")
            if cancel_on_error:
                logger.msg_error.emit("Cancel import on error is set. Bailing out.")
                if all_import_count > 0:
                    logger.msg_error.emit("Rolling back changes.")
                    client.call_method("rollback_session")
                    all_import_count = 0
                break
            logger.msg_warning.emit("Ignoring errors. Set Cancel import on error to bail out instead.")
    if all_import_count > 0:
        client.call_method("commit_session", "Import data by Spine Toolbox Importer")
        clean_url = clear_filter_configs(remove_credentials_from_url(client.get_db_url()))
        logger.msg_success.emit(
            f"Inserted {all_import_count} data with {len(all_import_errors)} errors into {clean_url}"
        )
    else:
        logger.msg_warning.emit("No new data imported")
    if all_import_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_import_error.log"))
        with open(logfilepath, "w", encoding="utf-8") as f:
            for err in all_import_errors:
                f.write(str(err) + "\n")
        # Make error log file anchor with path as tooltip
        logfile_anchor = (
            "<a style='color:#BB99FF;' title='" + logfilepath + "' href='file:///" + logfilepath + "'>Error log</a>"
        )
        logger.msg_error.emit(logfile_anchor)
        return False
    return True


def _parse_mappings(mapping_specs):
    parsed_mappings = {}
    for table, mapping_specs in mapping_specs.items():
        parsed_mappings[table] = [(parse_named_mapping_spec(spec)) for spec in mapping_specs]
    return parsed_mappings



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ToolSpecificationEditorWindow class and tool_spec_optional_widgets module."""
import logging
import os
from pathlib import Path
import sys
from tempfile import NamedTemporaryFile, TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtCore import QItemSelectionModel, Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication
from spine_items.tool.tool_specifications import ExecutableTool, JuliaTool, PythonTool
from spine_items.tool.widgets.tool_spec_optional_widgets import (
    ExecutableToolSpecOptionalWidget,
    JuliaToolSpecOptionalWidget,
    PythonToolSpecOptionalWidget,
)
from spine_items.tool.widgets.tool_specification_editor_window import ToolSpecificationEditorWindow
from tests.mock_helpers import MockQSettings, create_mock_toolbox_with_mock_qsettings


class TestToolSpecificationEditorWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Overridden method. Runs once before all tests in this class."""
        try:
            cls.app = QApplication().processEvents()
        except RuntimeError:
            pass
        logging.basicConfig(
            stream=sys.stderr,
            level=logging.DEBUG,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def setUp(self):
        """Overridden method. Runs before each test."""
        self._temp_dir = TemporaryDirectory()
        self.toolbox = create_mock_toolbox_with_mock_qsettings()
        self.tool_specification_widget = None

    def tearDown(self):
        """Overridden method. Runs after each test.
        Use this to free resources after a test if needed.
        """
        self._temp_dir.cleanup()
        if self.tool_specification_widget is not None:
            with mock.patch(
                "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase.tear_down"
            ) as mock_tear_down:
                mock_tear_down.return_value = True
                self.tool_specification_widget.close()
                mock_tear_down.assert_called()
            self.tool_specification_widget.deleteLater()
            self.tool_specification_widget = None

    def make_tool_spec_editor(self, spec=None):
        if not spec:
            with mock.patch("spinetoolbox.project_item.specification_editor_window.restore_ui") as mock_restore_ui:
                self.tool_specification_widget = ToolSpecificationEditorWindow(self.toolbox)
                mock_restore_ui.assert_called()
        else:
            with (
                mock.patch("spinetoolbox.project_item.specification_editor_window.restore_ui") as mock_restore_ui,
                mock.patch(
                    "spine_items.tool.tool_specifications.ToolSpecification._includes_main_path_relative"
                ) as mock_impr,
            ):
                mock_impr.return_value = ""
                self.tool_specification_widget = ToolSpecificationEditorWindow(self.toolbox, spec)
                mock_restore_ui.assert_called()
                mock_impr.assert_called()

    def test_change_tooltype(self):
        self.make_tool_spec_editor()
        self.tool_specification_widget._ui.comboBox_tooltype.setCurrentIndex(0)  # julia
        self.assertIsInstance(self.tool_specification_widget._get_optional_widget("julia"), JuliaToolSpecOptionalWidget)
        sd = self.tool_specification_widget.spec_dict
        exec_settings = sd.get("execution_settings")
        self.assertEqual(sd["tooltype"], "julia")
        self.assertEqual(len(exec_settings), 5)
        self.tool_specification_widget._ui.comboBox_tooltype.setCurrentIndex(1)  # python
        self.assertIsInstance(
            self.tool_specification_widget._get_optional_widget("python"), PythonToolSpecOptionalWidget
        )
        sd = self.tool_specification_widget.spec_dict
        exec_settings = sd.get("execution_settings")
        self.assertEqual(sd["tooltype"], "python")
        self.assertEqual(len(exec_settings), 4)
        self.tool_specification_widget._ui.comboBox_tooltype.setCurrentIndex(2)  # gams
        self.assertIsNone(self.tool_specification_widget._get_optional_widget("gams"))
        sd = self.tool_specification_widget.spec_dict
        exec_settings = sd.get("execution_settings")
        self.assertEqual(sd["tooltype"], "gams")
        self.assertIsNone(exec_settings)
        self.tool_specification_widget._ui.comboBox_tooltype.setCurrentIndex(3)  # executable
        self.assertIsInstance(
            self.tool_specification_widget._get_optional_widget("executable"), ExecutableToolSpecOptionalWidget
        )
        sd = self.tool_specification_widget.spec_dict
        exec_settings = sd.get("execution_settings")
        self.assertEqual(sd["tooltype"], "executable")
        self.assertEqual(len(exec_settings), 2)

    def test_make_new_specification(self):
        self.make_tool_spec_editor()
        self.tool_specification_widget._ui.comboBox_tooltype.setCurrentIndex(0)  # julia
        self.assertIsInstance(self.tool_specification_widget.optional_widget, JuliaToolSpecOptionalWidget)
        self.tool_specification_widget._spec_toolbar._line_edit_name.setText("test_julia_tool")
        with NamedTemporaryFile(mode="r") as temp_file:
            self.tool_specification_widget._set_main_program_file(temp_file.name)
            spec = self.tool_specification_widget._make_new_specification("test_julia_tool")
            self.tool_specification_widget._init_optional_widget(spec)
            self.assertIsInstance(spec, JuliaTool)
            self.assertEqual(len(spec.execution_settings), 5)
        self._call_save()

    def _call_save(self):
        """Calls tool spec widgets _save() while Toolbox's tool spec widget base _save() is mocked."""
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._save"
        ) as mock_save:
            mock_save.return_value = True
            self.tool_specification_widget._save()
            mock_save.assert_called()

    def test_open_tool_specification_editor_with_julia_spec(self):
        script_dir = Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = "hello.jl"
        file_path = Path(script_dir, script_file_name)
        with open(file_path, "w") as script_file:
            script_file.writelines(["println('hello')\n"])
        mock_logger = mock.MagicMock()
        julia_tool_spec = JuliaTool(
            "test_julia_spec",
            "julia",
            str(script_dir),
            [script_file_name],
            MockQSettings(),
            mock_logger,
            "Description",
            inputfiles=["data.csv"],
            inputfiles_opt=["*.dat"],
            outputfiles=["results.txt"],
            cmdline_args=["-A", "-B"],
        )
        julia_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(julia_tool_spec)
        opt_widget = self.tool_specification_widget.optional_widget
        self.assertIsInstance(opt_widget, JuliaToolSpecOptionalWidget)
        self.assertTrue(self.tool_specification_widget._ui.comboBox_tooltype.currentText() == "Julia")
        self.assertTrue(self.tool_specification_widget._ui.lineEdit_args.text() == "-A -B")
        # Program files dock widget should have 2 rows :'Main program file' and 'Additional program files'
        self.assertEqual(2, self.tool_specification_widget.programfiles_model.rowCount())
        # Get index of 'Main program file' item
        parent = self.tool_specification_widget.programfiles_model.index(0, 0)
        # There should be one row under 'Main program file' -> the main program file
        self.assertEqual(1, self.tool_specification_widget.programfiles_model.rowCount(parent))
        index = self.tool_specification_widget.programfiles_model.index(0, 0, parent)  # Index of 'hello.jl'
        item = self.tool_specification_widget.programfiles_model.itemFromIndex(index)
        self.assertEqual(script_file_name, item.data(Qt.ItemDataRole.DisplayRole))
        # Check Input & output files dock widget
        self.assertEqual(3, self.tool_specification_widget.io_files_model.rowCount())
        if_index = self.tool_specification_widget.io_files_model.index(0, 0)  # 'Input files' item index
        oif_index = self.tool_specification_widget.io_files_model.index(1, 0)  # 'Optional input files' item index
        of_index = self.tool_specification_widget.io_files_model.index(2, 0)  # 'Output files' item index
        if_child_index = self.tool_specification_widget.io_files_model.index(0, 0, if_index)
        oif_child_index = self.tool_specification_widget.io_files_model.index(0, 0, oif_index)
        of_child_index = self.tool_specification_widget.io_files_model.index(0, 0, of_index)
        self.assertEqual(1, self.tool_specification_widget.io_files_model.rowCount(if_index))
        self.assertEqual(1, self.tool_specification_widget.io_files_model.rowCount(oif_index))
        self.assertEqual(1, self.tool_specification_widget.io_files_model.rowCount(of_index))
        if_item = self.tool_specification_widget.io_files_model.itemFromIndex(if_child_index)
        self.assertEqual("data.csv", if_item.data(Qt.ItemDataRole.DisplayRole))
        oif_item = self.tool_specification_widget.io_files_model.itemFromIndex(oif_child_index)
        self.assertEqual("*.dat", oif_item.data(Qt.ItemDataRole.DisplayRole))
        of_item = self.tool_specification_widget.io_files_model.itemFromIndex(of_child_index)
        self.assertEqual("results.txt", of_item.data(Qt.ItemDataRole.DisplayRole))

    def test_open_tool_specification_editor_with_executable_spec(self):
        mock_logger = mock.MagicMock()
        exec_tool_spec = ExecutableTool(
            "a", "executable", self._temp_dir.name, ["fake_main_program.bat"], MockQSettings(), mock_logger
        )
        exec_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(exec_tool_spec)
        opt_widget = self.tool_specification_widget._get_optional_widget("executable")
        self.assertIsInstance(opt_widget, ExecutableToolSpecOptionalWidget)
        self.assertFalse(opt_widget.ui.lineEdit_command.isEnabled())  # Command is disabled when a program file is set
        # Program files dock widet should have 2 rows :'Main program file' and 'Additional program files'
        self.assertEqual(2, self.tool_specification_widget.programfiles_model.rowCount())
        # Get index of 'Main program file' item
        parent = self.tool_specification_widget.programfiles_model.index(0, 0)
        # There should be one row under 'Main program file' -> the main program file
        self.assertEqual(1, self.tool_specification_widget.programfiles_model.rowCount(parent))
        index = self.tool_specification_widget.programfiles_model.index(
            0, 0, parent
        )  # Index of 'fake_main_program.bat'
        item = self.tool_specification_widget.programfiles_model.itemFromIndex(index)
        self.assertEqual("fake_main_program.bat", item.data(Qt.ItemDataRole.DisplayRole))

    def test_edit_and_save_program_file(self):
        mock_logger = mock.MagicMock()
        script_file_name = "hello.py"
        file_path = Path(self._temp_dir.name, script_file_name)
        with open(file_path, "w") as h:
            h.writelines(["# hello.py"])  # Make hello.py
        python_tool_spec = PythonTool(
            "a", "python", self._temp_dir.name, [script_file_name], MockQSettings(), mock_logger
        )
        python_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(python_tool_spec)
        self.tool_specification_widget._ui.textEdit_program.appendPlainText("print('hi')")
        self.tool_specification_widget._save_program_file(
            file_path, self.tool_specification_widget._ui.textEdit_program.document()
        )
        # Open file and check contents
        with open(file_path, "r") as edited_file:
            l = edited_file.readlines()
            self.assertEqual(2, len(l))
            self.assertTrue(l[0].startswith("# hello"))  # Don't match the whole str to avoid problems with newline
            self.assertTrue(l[1].startswith("print('hi')"))

    def test_change_python_spec_options(self):
        mock_logger = mock.MagicMock()
        script_file_name = "hello.py"
        file_path = Path(self._temp_dir.name, script_file_name)
        with open(file_path, "w") as h:
            h.writelines(["# hello.py"])  # Make hello.py
        python_tool_spec = PythonTool(
            "a", "python", self._temp_dir.name, [script_file_name], MockQSettings(), mock_logger
        )
        python_tool_spec.init_execution_settings()  # Sets defaults
        python_tool_spec.execution_settings["use_jupyter_console"] = True
        python_tool_spec.execution_settings["kernel_spec_name"] = "python310"
        with mock.patch("spine_items.tool.widgets.tool_spec_optional_widgets.KernelFetcher", new=FakeKernelFetcher):
            self.make_tool_spec_editor(python_tool_spec)
            opt_widget = self.tool_specification_widget.optional_widget
            self.assertTrue(opt_widget.ui.radioButton_jupyter_console.isChecked())
            self.assertEqual(3, opt_widget.kernel_spec_model.rowCount())
            self.assertEqual(1, opt_widget.ui.comboBox_kernel_specs.currentIndex())
            self.assertEqual("python310", opt_widget.ui.comboBox_kernel_specs.currentText())
            self.tool_specification_widget.push_change_kernel_spec_command(2)
            self.assertEqual(
                "python311", self.tool_specification_widget.spec_dict["execution_settings"]["kernel_spec_name"]
            )
            self.assertEqual(2, opt_widget.ui.comboBox_kernel_specs.currentIndex())
            self.assertEqual("python311", opt_widget.ui.comboBox_kernel_specs.currentText())
            self.assertTrue(self.tool_specification_widget.spec_dict["execution_settings"]["use_jupyter_console"])
            # Test SharedToolSpecOptionalWidget._restore_selected_kernel()
            # Restore selected kernel after the kernel spec model has been reloaded
            opt_widget.start_kernel_fetcher()
            self.assertEqual(3, opt_widget.kernel_spec_model.rowCount())
            self.assertEqual(2, opt_widget.ui.comboBox_kernel_specs.currentIndex())
            self.assertEqual("python311", opt_widget.ui.comboBox_kernel_specs.currentText())
            # Test push_set_jupyter_console_mode()
            self.tool_specification_widget.push_set_jupyter_console_mode(False)
            self.assertFalse(self.tool_specification_widget.spec_dict["execution_settings"]["use_jupyter_console"])
            opt_widget.set_executable("path/to/executable")
            self.tool_specification_widget.push_change_executable(opt_widget.get_executable())
            self.assertEqual(
                "path/to/executable", self.tool_specification_widget.spec_dict["execution_settings"]["executable"]
            )
            self.tool_specification_widget._push_change_args_command("-A -B")
            self.assertEqual(["-A", "-B"], self.tool_specification_widget.spec_dict["cmdline_args"])

    def test_change_executable_spec_options(self):
        mock_logger = mock.MagicMock()
        batch_file = "hello.bat"
        another_batch_file = "hello.sh"
        exec_tool_spec = ExecutableTool(
            "a", "executable", self._temp_dir.name, [batch_file, "data.file"], MockQSettings(), mock_logger
        )
        exec_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(exec_tool_spec)
        file_path = Path(self._temp_dir.name, another_batch_file)
        parent_main = self.tool_specification_widget.programfiles_model.index(0, 0)  # Main program file item
        parent_addit = self.tool_specification_widget.programfiles_model.index(1, 0)  # Additional program files item
        mp_index = self.tool_specification_widget.programfiles_model.index(0, 0, parent_main)
        # Try to open main program file
        with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.open_url") as mock_open_url:
            self.tool_specification_widget.open_program_file(mp_index)  # Fails because it's a .bat
            mock_open_url.assert_not_called()
        # Change main program file
        self.tool_specification_widget._push_change_main_program_file_command(file_path)
        mp_index = self.tool_specification_widget.programfiles_model.index(0, 0, parent_main)
        # Try to open main program file again
        with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.open_url") as mock_open_url:
            self.tool_specification_widget.open_program_file(mp_index)  # Calls open_url()
            mock_open_url.assert_called()
        # Try removing files without selecting anything
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._show_status_bar_msg"
        ) as m_notify:
            self.tool_specification_widget.remove_program_files()
            m_notify.assert_called()
        # Set 'data.file' selected
        self.assertEqual(1, self.tool_specification_widget.programfiles_model.rowCount(parent_main))
        self.assertEqual(1, self.tool_specification_widget.programfiles_model.rowCount(parent_addit))
        index = self.tool_specification_widget.programfiles_model.index(0, 0, parent_addit)  # Index of 'data.file'
        selection_model = self.tool_specification_widget._ui.treeView_programfiles.selectionModel()
        selection_model.setCurrentIndex(index, QItemSelectionModel.SelectionFlag.Select)
        # Remove additional program file 'data.file'
        self.tool_specification_widget.remove_program_files()
        self.assertEqual(0, self.tool_specification_widget.programfiles_model.rowCount(parent_addit))
        # Remove main program file
        self.tool_specification_widget.remove_all_program_files()
        self.assertEqual(0, self.tool_specification_widget.programfiles_model.rowCount(parent_main))
        # Do remove_all without selecting anything
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._show_status_bar_msg"
        ) as m_notify:
            self.tool_specification_widget.remove_all_program_files()
            m_notify.assert_called()
        # Add command for executable tool spec
        self.tool_specification_widget.optional_widget.ui.lineEdit_command.setText("ls -a")
        self.tool_specification_widget.push_change_executable_command(
            self.tool_specification_widget.optional_widget.ui.lineEdit_command.text()
        )
        # Check that no shell is selected
        self.assertEqual("", self.tool_specification_widget.optional_widget.get_current_shell())
        self.assertEqual("No shell", self.tool_specification_widget.optional_widget.ui.comboBox_shell.currentText())
        # Change shell to cmd.exe on win32, bash for others
        if sys.platform == "win32":
            index_of_cmd_exe = self.tool_specification_widget.optional_widget.shells.index("cmd.exe")
            self.tool_specification_widget.push_change_shell_command(index_of_cmd_exe)
            self.assertEqual("cmd.exe", self.tool_specification_widget.optional_widget.get_current_shell())
        else:
            index_of_bash = self.tool_specification_widget.optional_widget.shells.index("bash")
            self.tool_specification_widget.push_change_shell_command(index_of_bash)
            self.assertEqual("bash", self.tool_specification_widget.optional_widget.get_current_shell())

    def test_change_julia_project(self):
        mock_logger = mock.MagicMock()
        julia_tool_spec = JuliaTool("a", "julia", self._temp_dir.name, ["hello.jl"], MockQSettings(), mock_logger)
        julia_tool_spec.init_execution_settings()  # Sets defaults
        julia_tool_spec.execution_settings["use_jupyter_console"] = True
        with mock.patch("spine_items.tool.widgets.tool_spec_optional_widgets.KernelFetcher", new=FakeKernelFetcher):
            self.make_tool_spec_editor(julia_tool_spec)
            self.assertEqual("", self.tool_specification_widget.spec_dict["execution_settings"]["project"])
            self.tool_specification_widget.optional_widget.ui.lineEdit_julia_project.setText("path/to/julia_project")
            self.tool_specification_widget.push_change_project()
            self.assertEqual(
                "path/to/julia_project", self.tool_specification_widget.spec_dict["execution_settings"]["project"]
            )

    def test_restore_unknown_saved_kernel_into_optional_widget(self):
        mock_logger = mock.MagicMock()
        script_file_name = "hello.py"
        file_path = Path(self._temp_dir.name, script_file_name)
        with open(file_path, "w") as h:
            h.writelines(["# hello.py"])  # Make hello.py
        python_tool_spec = PythonTool(
            "a", "python", self._temp_dir.name, [script_file_name], MockQSettings(), mock_logger
        )
        python_tool_spec.init_execution_settings()  # Sets defaults
        python_tool_spec.execution_settings["use_jupyter_console"] = True
        python_tool_spec.execution_settings["kernel_spec_name"] = "unknown_kernel"
        with (
            mock.patch("spine_items.tool.widgets.tool_spec_optional_widgets.KernelFetcher", new=FakeKernelFetcher),
            mock.patch("spine_items.tool.widgets.tool_spec_optional_widgets.Notification") as mock_notify,
        ):
            self.make_tool_spec_editor(python_tool_spec)
            opt_widget = self.tool_specification_widget.optional_widget
            self.assertTrue(opt_widget.ui.radioButton_jupyter_console.isChecked())
            self.assertEqual(3, opt_widget.kernel_spec_model.rowCount())
            self.assertEqual(0, opt_widget.ui.comboBox_kernel_specs.currentIndex())
            self.assertEqual("Select kernel spec...", opt_widget.ui.comboBox_kernel_specs.currentText())
            mock_notify.assert_called()

    def test_program_file_dialogs(self):
        mock_logger = mock.MagicMock()
        script_file_name = "hello.jl"
        script_file_name2 = "hello2.jl"
        script_file_name3 = "hello3.jl"
        data_file_name = "data.csv"
        file_path = Path(self._temp_dir.name, script_file_name)
        file_path2 = Path(self._temp_dir.name, script_file_name2)
        file_path3 = Path(self._temp_dir.name, data_file_name)
        file_path4 = Path(self._temp_dir.name, script_file_name3)
        # Make files so os.path.samefile() works
        with open(file_path, "w") as h:
            h.writelines(["println('Hello world')"])  # Make hello.jl
        with open(file_path2, "w") as h:
            h.writelines(["println('Hello world2')"])  # Make hello2.jl
        with open(file_path3, "w") as h:
            h.writelines(["1, 2, 3"])  # Make data.csv
        julia_tool_spec = JuliaTool(
            "a", "julia", self._temp_dir.name, [script_file_name, data_file_name], MockQSettings(), mock_logger
        )
        julia_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(julia_tool_spec)
        self.assertEqual("hello.jl", os.path.split(self.tool_specification_widget._current_main_program_file())[1])
        # Test browse_main_program_file()
        with mock.patch(
            "spine_items.tool.widgets.tool_specification_editor_window.QFileDialog.getOpenFileName"
        ) as mock_fd_gofn:
            mock_fd_gofn.return_value = [file_path2]
            # Change main program file hello.jl -> hello2.jl
            self.tool_specification_widget.browse_main_program_file()
            mock_fd_gofn.assert_called()
            self.assertEqual("hello2.jl", os.path.split(self.tool_specification_widget._current_main_program_file())[1])
            # Try to change additional program file as the main program, should pop up a QMessageBox
            mock_fd_gofn.return_value = [file_path3]
            with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.QMessageBox") as mock_mb:
                self.tool_specification_widget.browse_main_program_file()
                mock_mb.assert_called()
            self.assertEqual("hello2.jl", os.path.split(self.tool_specification_widget._current_main_program_file())[1])
            # Rename the current main program file hello2.jl -> hello3.jl
            os.rename(file_path2, file_path4)
            mock_fd_gofn.return_value = [file_path4]
            # Select renamed main program file hello3.jl
            self.tool_specification_widget.browse_main_program_file()
            mock_fd_gofn.assert_called()
            self.assertEqual("hello3.jl", os.path.split(self.tool_specification_widget._current_main_program_file())[1])
        # Test new_main_program_file()
        with mock.patch(
            "spine_items.tool.widgets.tool_specification_editor_window.QFileDialog.getSaveFileName"
        ) as mock_fd_gsfn:
            mock_fd_gsfn.return_value = [file_path]
            # This should remove existing hello.jl, recreate it, and set it as main program
            self.tool_specification_widget.new_main_program_file()
            self.assertEqual(1, mock_fd_gsfn.call_count)
            self.assertEqual("hello.jl", os.path.split(self.tool_specification_widget._current_main_program_file())[1])
            # Test new_program_file()
            parent_addit = self.tool_specification_widget.programfiles_model.index(1, 0)
            self.assertEqual(1, self.tool_specification_widget.programfiles_model.rowCount(parent_addit))
            mock_fd_gsfn.return_value = [Path(self._temp_dir.name, "input.txt")]
            self.tool_specification_widget.new_program_file()
            self.assertEqual(2, mock_fd_gsfn.call_count)
            # Check that we now have 2 additional program files
            parent_addit = self.tool_specification_widget.programfiles_model.index(1, 0)
            self.assertEqual(2, self.tool_specification_widget.programfiles_model.rowCount(parent_addit))
            # Try to add file that's already been added
            mock_fd_gsfn.return_value = [file_path3]
            with mock.patch(
                "spine_items.tool.widgets.tool_specification_editor_window.QMessageBox.information"
            ) as mock_mb_info:
                self.tool_specification_widget.new_program_file()  # QMessageBox should appear
                mock_mb_info.assert_called()
            self.assertEqual(3, mock_fd_gsfn.call_count)
            self.assertEqual(2, self.tool_specification_widget.programfiles_model.rowCount(parent_addit))
        # Test show_add_program_files_dialog()
        with mock.patch(
            "spine_items.tool.widgets.tool_specification_editor_window.QFileDialog.getOpenFileNames"
        ) as mock_fd_gofns:
            with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.QMessageBox") as mock_mb:
                mock_fd_gofns.return_value = [[file_path]]
                self.tool_specification_widget.show_add_program_files_dialog()  # Shows 'Can't add main...' msg box
                self.assertEqual(1, mock_mb.call_count)
                self.assertEqual(1, mock_fd_gofns.call_count)
                mock_fd_gofns.return_value = [[file_path, file_path2, file_path3]]
                self.tool_specification_widget.show_add_program_files_dialog()  # Shows 'One file not add...' msg box
                self.assertEqual(2, mock_mb.call_count)
                self.assertEqual(2, mock_fd_gofns.call_count)
        # Test show_add_program_dirs_dialog()
        with mock.patch(
            "spine_items.tool.widgets.tool_specification_editor_window.QFileDialog.getExistingDirectory"
        ) as mock_fd_ged:
            mock_fd_ged.return_value = self._temp_dir.name
            self.tool_specification_widget.show_add_program_dirs_dialog()
            mock_fd_ged.assert_called()

    def test_add_rename_select_remove_input_and_output_files(self):
        mock_logger = mock.MagicMock()
        main_file = "hello.jl"
        main_path = Path(self._temp_dir.name, main_file)
        with open(main_path, "w") as h:
            h.writelines(["println('Hello world')"])  # Make hello.jl
        julia_tool_spec = JuliaTool("a", "julia", self._temp_dir.name, [main_file], MockQSettings(), mock_logger)
        julia_tool_spec.init_execution_settings()  # Sets defaults
        self.make_tool_spec_editor(julia_tool_spec)
        # INPUT FILES
        # Test add_input_files()
        with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.QInputDialog.getText") as mock_gt:
            mock_gt.return_value = ["data.csv"]
            self.tool_specification_widget.add_inputfiles()
            mock_gt.assert_called()
        iofm = self.tool_specification_widget.io_files_model
        selection_model = self.tool_specification_widget._ui.treeView_io_files.selectionModel()
        self.assertEqual(3, iofm.rowCount())  # row 0: Input files, row 1: Optional input files, row 2: Output files
        # There should be one item under 'Input files'
        input_files_root = iofm.index(0, 0)
        self.assertEqual(1, iofm.rowCount(input_files_root))
        input_file_item = iofm.itemFromIndex(iofm.index(0, 0, input_files_root))
        self.assertEqual("data.csv", input_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Test rename input file (tests _push_io_file_renamed_command())
        input_file_item.setData("renamed_file.csv", Qt.ItemDataRole.DisplayRole)
        input_file_item = iofm.itemFromIndex(iofm.index(0, 0, input_files_root))
        self.assertEqual("renamed_file.csv", input_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Try remove_inputfiles() without selections
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._show_status_bar_msg"
        ) as m_notify:
            self.tool_specification_widget.remove_inputfiles()
            m_notify.assert_called()
        # Select input file item
        selection_model.setCurrentIndex(iofm.index(0, 0, input_files_root), QItemSelectionModel.SelectionFlag.Select)
        # Test remove_inputfiles()
        self.tool_specification_widget.remove_inputfiles()
        self.assertEqual(0, iofm.rowCount(input_files_root))
        # OPTIONAL INPUT FILES
        # Test add_inputfiles_opt()
        with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.QInputDialog.getText") as mock_gt:
            mock_gt.return_value = ["*.dat"]
            self.tool_specification_widget.add_inputfiles_opt()
            mock_gt.assert_called()
        # There should be one item under 'Optional input files'
        opt_input_files_root = iofm.index(1, 0)
        self.assertEqual(1, iofm.rowCount(opt_input_files_root))
        opt_input_file_item = iofm.itemFromIndex(iofm.index(0, 0, opt_input_files_root))
        self.assertEqual("*.dat", opt_input_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Test rename optional input item (tests _push_io_file_renamed_command())
        opt_input_file_item.setData("???.dat", Qt.ItemDataRole.DisplayRole)
        opt_input_file_item = iofm.itemFromIndex(iofm.index(0, 0, opt_input_files_root))
        self.assertEqual("???.dat", opt_input_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Try remove_inputfiles_opt() without selections
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._show_status_bar_msg"
        ) as m_notify:
            self.tool_specification_widget.remove_inputfiles_opt()
            m_notify.assert_called()
        # Select optional input file item
        selection_model.setCurrentIndex(
            iofm.index(0, 0, opt_input_files_root), QItemSelectionModel.SelectionFlag.Select
        )
        # Test remove_inputfiles_opt()
        self.tool_specification_widget.remove_inputfiles_opt()
        self.assertEqual(0, iofm.rowCount(opt_input_files_root))
        # OUTPUT FILES
        # Test add_outputfiles()
        with mock.patch("spine_items.tool.widgets.tool_specification_editor_window.QInputDialog.getText") as mock_gt:
            mock_gt.return_value = ["results.txt"]
            self.tool_specification_widget.add_outputfiles()
            mock_gt.assert_called()
        # There should be one item under 'Output files'
        output_files_root = iofm.index(2, 0)
        self.assertEqual(1, iofm.rowCount(output_files_root))
        output_file_item = iofm.itemFromIndex(iofm.index(0, 0, output_files_root))
        self.assertEqual("results.txt", output_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Test rename output file
        output_file_item.setData("output.txt", Qt.ItemDataRole.DisplayRole)
        output_file_item = iofm.itemFromIndex(iofm.index(0, 0, output_files_root))
        self.assertEqual("output.txt", output_file_item.data(Qt.ItemDataRole.DisplayRole))
        # Try remove_outputfiles() without selections
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase._show_status_bar_msg"
        ) as m_notify:
            self.tool_specification_widget.remove_outputfiles()
            m_notify.assert_called()
        # Select output file item
        selection_model.setCurrentIndex(iofm.index(0, 0, output_files_root), QItemSelectionModel.SelectionFlag.Select)
        # Test remove_outputfiles()
        self.tool_specification_widget.remove_outputfiles()
        self.assertEqual(0, iofm.rowCount(output_files_root))


class FakeSignal:
    def __init__(self):
        self.call_list = []  # List of slots

    def connect(self, method):
        """Stores all slots connected to this FakeSignal into a list."""
        self.call_list.append(method)


class FakeKernelFetcher:
    """Class for replacing KernelFetcher in tests."""

    kernel_found = FakeSignal()
    finished = FakeSignal()

    def __init__(self, conda_path="", fetch_mode=0):
        self.conda_path = conda_path
        self.fetch_mode = fetch_mode

    def isRunning(self):
        return False

    def start(self):
        for m in self.kernel_found.call_list:
            # Calls SharedToolSpecOptionalWidget.add_kernel()
            m("python310", "", False, QIcon(), {})
            m("python311", "", False, QIcon(), {})
        for meth in self.finished.call_list:
            # Calls two methods:
            # 1. Either SharedToolSpecOptionalWidget._restore_saved_kernel() or
            # SharedToolSpecOptionalWidget._restore_selected_kernel()
            # 2. mock.restore_overrider_cursor()
            # Note: The order of connect() calls matters.
            meth()
        # Clear signal slots, so this can be used again
        self.kernel_found.call_list.clear()
        self.finished.call_list.clear()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'import_editor_window.ui'
##
## Created by: Qt User Interface Compiler version 6.8.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,
    QFrame, QGridLayout, QHBoxLayout, QHeaderView,
    QLabel, QLineEdit, QListView, QMainWindow,
    QPushButton, QSizePolicy, QSpacerItem, QSpinBox,
    QSplitter, QTableView, QToolButton, QVBoxLayout,
    QWidget)

from spine_items.importer.widgets.multi_checkable_list_view import MultiCheckableListView
from spine_items.importer.widgets.table_view_with_button_header import TableViewWithButtonHeader
from spine_items import resources_icons_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(929, 732)
        MainWindow.setDockNestingEnabled(True)
        self.export_mappings_action = QAction(MainWindow)
        self.export_mappings_action.setObjectName(u"export_mappings_action")
        self.export_mappings_action.setEnabled(False)
        self.import_mappings_action = QAction(MainWindow)
        self.import_mappings_action.setObjectName(u"import_mappings_action")
        self.import_mappings_action.setEnabled(False)
        self.switch_input_type_action = QAction(MainWindow)
        self.switch_input_type_action.setObjectName(u"switch_input_type_action")
        self.switch_input_type_action.setEnabled(False)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_6 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.input_path_label = QLabel(self.centralwidget)
        self.input_path_label.setObjectName(u"input_path_label")

        self.horizontalLayout_2.addWidget(self.input_path_label)

        self.input_path_line_edit = QLineEdit(self.centralwidget)
        self.input_path_line_edit.setObjectName(u"input_path_line_edit")
        self.input_path_line_edit.setClearButtonEnabled(True)

        self.horizontalLayout_2.addWidget(self.input_path_line_edit)

        self.browse_inputs_button = QToolButton(self.centralwidget)
        self.browse_inputs_button.setObjectName(u"browse_inputs_button")
        icon = QIcon()
        icon.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.browse_inputs_button.setIcon(icon)

        self.horizontalLayout_2.addWidget(self.browse_inputs_button)

        self.input_type_label = QLabel(self.centralwidget)
        self.input_type_label.setObjectName(u"input_type_label")

        self.horizontalLayout_2.addWidget(self.input_type_label)

        self.input_type_line_edit = QLineEdit(self.centralwidget)
        self.input_type_line_edit.setObjectName(u"input_type_line_edit")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.input_type_line_edit.sizePolicy().hasHeightForWidth())
        self.input_type_line_edit.setSizePolicy(sizePolicy)
        self.input_type_line_edit.setReadOnly(True)

        self.horizontalLayout_2.addWidget(self.input_type_line_edit)


        self.verticalLayout_6.addLayout(self.horizontalLayout_2)

        self.splitter_source_list = QSplitter(self.centralwidget)
        self.splitter_source_list.setObjectName(u"splitter_source_list")
        self.splitter_source_list.setOrientation(Qt.Orientation.Horizontal)
        self.source_list = MultiCheckableListView(self.splitter_source_list)
        self.source_list.setObjectName(u"source_list")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.source_list.sizePolicy().hasHeightForWidth())
        self.source_list.setSizePolicy(sizePolicy1)
        self.source_list.setMaximumSize(QSize(16777215, 16777215))
        self.source_list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.source_list.setAcceptDrops(True)
        self.source_list.setDragEnabled(True)
        self.source_list.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.source_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.source_list.setTextElideMode(Qt.TextElideMode.ElideLeft)
        self.splitter_source_list.addWidget(self.source_list)
        self.splitter_source_data_mappings = QSplitter(self.splitter_source_list)
        self.splitter_source_data_mappings.setObjectName(u"splitter_source_data_mappings")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(5)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.splitter_source_data_mappings.sizePolicy().hasHeightForWidth())
        self.splitter_source_data_mappings.setSizePolicy(sizePolicy2)
        self.splitter_source_data_mappings.setOrientation(Qt.Orientation.Horizontal)
        self.frame_source_data = QFrame(self.splitter_source_data_mappings)
        self.frame_source_data.setObjectName(u"frame_source_data")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.frame_source_data.sizePolicy().hasHeightForWidth())
        self.frame_source_data.setSizePolicy(sizePolicy3)
        self.frame_source_data.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_source_data.setFrameShadow(QFrame.Shadow.Raised)
        self.frame_source_data.setLineWidth(0)
        self.frame_source_data.setMidLineWidth(0)
        self.verticalLayout = QVBoxLayout(self.frame_source_data)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_source_options = QVBoxLayout()
        self.verticalLayout_source_options.setSpacing(0)
        self.verticalLayout_source_options.setObjectName(u"verticalLayout_source_options")

        self.verticalLayout.addLayout(self.verticalLayout_source_options)

        self.source_data_table = TableViewWithButtonHeader(self.frame_source_data)
        self.source_data_table.setObjectName(u"source_data_table")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.source_data_table.sizePolicy().hasHeightForWidth())
        self.source_data_table.setSizePolicy(sizePolicy4)
        self.source_data_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self.verticalLayout.addWidget(self.source_data_table)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(1, 1, 1, 1)
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.label = QLabel(self.frame_source_data)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.default_column_type_combo_box = QComboBox(self.frame_source_data)
        self.default_column_type_combo_box.setObjectName(u"default_column_type_combo_box")

        self.horizontalLayout.addWidget(self.default_column_type_combo_box)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.splitter_source_data_mappings.addWidget(self.frame_source_data)
        self.splitter = QSplitter(self.splitter_source_data_mappings)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.verticalLayoutWidget_2 = QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName(u"verticalLayoutWidget_2")
        self.verticalLayout_4 = QVBoxLayout(self.verticalLayoutWidget_2)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.label_2 = QLabel(self.verticalLayoutWidget_2)
        self.label_2.setObjectName(u"label_2")

        self.verticalLayout_4.addWidget(self.label_2)

        self.mapping_list = QListView(self.verticalLayoutWidget_2)
        self.mapping_list.setObjectName(u"mapping_list")
        self.mapping_list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.mapping_list.setAcceptDrops(True)
        self.mapping_list.setDragEnabled(True)
        self.mapping_list.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.mapping_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.verticalLayout_4.addWidget(self.mapping_list)

        self.splitter.addWidget(self.verticalLayoutWidget_2)
        self.frame_2 = QFrame(self.splitter)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.frame_2.setLineWidth(0)
        self.verticalLayout_2 = QVBoxLayout(self.frame_2)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(2, 2, 2, 2)
        self.button_layout = QHBoxLayout()
        self.button_layout.setObjectName(u"button_layout")
        self.button_layout.setContentsMargins(0, 0, 0, 0)
        self.new_button = QPushButton(self.frame_2)
        self.new_button.setObjectName(u"new_button")

        self.button_layout.addWidget(self.new_button)

        self.remove_button = QPushButton(self.frame_2)
        self.remove_button.setObjectName(u"remove_button")

        self.button_layout.addWidget(self.remove_button)

        self.duplicate_button = QPushButton(self.frame_2)
        self.duplicate_button.setObjectName(u"duplicate_button")

        self.button_layout.addWidget(self.duplicate_button)


        self.verticalLayout_2.addLayout(self.button_layout)

        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.mapping_options_contents = QGridLayout()
        self.mapping_options_contents.setSpacing(3)
        self.mapping_options_contents.setObjectName(u"mapping_options_contents")
        self.mapping_options_contents.setContentsMargins(3, 3, 3, 3)
        self.class_type_combo_box = QComboBox(self.frame_2)
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.addItem("")
        self.class_type_combo_box.setObjectName(u"class_type_combo_box")

        self.mapping_options_contents.addWidget(self.class_type_combo_box, 0, 1, 1, 1)

        self.value_type_label = QLabel(self.frame_2)
        self.value_type_label.setObjectName(u"value_type_label")

        self.mapping_options_contents.addWidget(self.value_type_label, 3, 0, 1, 1)

        self.value_type_combo_box = QComboBox(self.frame_2)
        self.value_type_combo_box.addItem("")
        self.value_type_combo_box.addItem("")
        self.value_type_combo_box.addItem("")
        self.value_type_combo_box.addItem("")
        self.value_type_combo_box.addItem("")
        self.value_type_combo_box.setObjectName(u"value_type_combo_box")

        self.mapping_options_contents.addWidget(self.value_type_combo_box, 3, 1, 1, 1)

        self.start_read_row_spin_box = QSpinBox(self.frame_2)
        self.start_read_row_spin_box.setObjectName(u"start_read_row_spin_box")
        self.start_read_row_spin_box.setMinimum(1)

        self.mapping_options_contents.addWidget(self.start_read_row_spin_box, 7, 1, 1, 1)

        self.class_type_label = QLabel(self.frame_2)
        self.class_type_label.setObjectName(u"class_type_label")

        self.mapping_options_contents.addWidget(self.class_type_label, 0, 0, 1, 1)

        self.dimension_spin_box = QSpinBox(self.frame_2)
        self.dimension_spin_box.setObjectName(u"dimension_spin_box")
        self.dimension_spin_box.setValue(0)

        self.mapping_options_contents.addWidget(self.dimension_spin_box, 4, 1, 1, 1)

        self.parameter_type_label = QLabel(self.frame_2)
        self.parameter_type_label.setObjectName(u"parameter_type_label")

        self.mapping_options_contents.addWidget(self.parameter_type_label, 2, 0, 1, 1)

        self.dimension_label = QLabel(self.frame_2)
        self.dimension_label.setObjectName(u"dimension_label")

        self.mapping_options_contents.addWidget(self.dimension_label, 4, 0, 1, 1)

        self.ignore_columns_button = QPushButton(self.frame_2)
        self.ignore_columns_button.setObjectName(u"ignore_columns_button")

        self.mapping_options_contents.addWidget(self.ignore_columns_button, 8, 1, 1, 1)

        self.import_entity_alternatives_check_box = QCheckBox(self.frame_2)
        self.import_entity_alternatives_check_box.setObjectName(u"import_entity_alternatives_check_box")

        self.mapping_options_contents.addWidget(self.import_entity_alternatives_check_box, 9, 0, 1, 1)

        self.map_dimension_spin_box = QSpinBox(self.frame_2)
        self.map_dimension_spin_box.setObjectName(u"map_dimension_spin_box")
        self.map_dimension_spin_box.setMinimum(1)

        self.mapping_options_contents.addWidget(self.map_dimension_spin_box, 5, 1, 1, 1)

        self.parameter_type_combo_box = QComboBox(self.frame_2)
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.setObjectName(u"parameter_type_combo_box")

        self.mapping_options_contents.addWidget(self.parameter_type_combo_box, 2, 1, 1, 1)

        self.read_start_row_label = QLabel(self.frame_2)
        self.read_start_row_label.setObjectName(u"read_start_row_label")

        self.mapping_options_contents.addWidget(self.read_start_row_label, 7, 0, 1, 1)

        self.ignore_columns_label = QLabel(self.frame_2)
        self.ignore_columns_label.setObjectName(u"ignore_columns_label")

        self.mapping_options_contents.addWidget(self.ignore_columns_label, 8, 0, 1, 1)

        self.map_dimensions_label = QLabel(self.frame_2)
        self.map_dimensions_label.setObjectName(u"map_dimensions_label")

        self.mapping_options_contents.addWidget(self.map_dimensions_label, 5, 0, 1, 1)

        self.import_entities_check_box = QCheckBox(self.frame_2)
        self.import_entities_check_box.setObjectName(u"import_entities_check_box")

        self.mapping_options_contents.addWidget(self.import_entities_check_box, 10, 0, 1, 1)

        self.time_series_repeat_check_box = QCheckBox(self.frame_2)
        self.time_series_repeat_check_box.setObjectName(u"time_series_repeat_check_box")

        self.mapping_options_contents.addWidget(self.time_series_repeat_check_box, 9, 1, 1, 1)

        self.map_compression_check_box = QCheckBox(self.frame_2)
        self.map_compression_check_box.setObjectName(u"map_compression_check_box")

        self.mapping_options_contents.addWidget(self.map_compression_check_box, 10, 1, 1, 1)


        self.verticalLayout_3.addLayout(self.mapping_options_contents)

        self.mapping_spec_table = QTableView(self.frame_2)
        self.mapping_spec_table.setObjectName(u"mapping_spec_table")
        self.mapping_spec_table.setLineWidth(0)
        self.mapping_spec_table.horizontalHeader().setStretchLastSection(True)

        self.verticalLayout_3.addWidget(self.mapping_spec_table)


        self.verticalLayout_2.addLayout(self.verticalLayout_3)

        self.splitter.addWidget(self.frame_2)
        self.splitter_source_data_mappings.addWidget(self.splitter)
        self.splitter_source_list.addWidget(self.splitter_source_data_mappings)

        self.verticalLayout_6.addWidget(self.splitter_source_list)

        MainWindow.setCentralWidget(self.centralwidget)
        QWidget.setTabOrder(self.input_path_line_edit, self.browse_inputs_button)
        QWidget.setTabOrder(self.browse_inputs_button, self.input_type_line_edit)
        QWidget.setTabOrder(self.input_type_line_edit, self.source_list)
        QWidget.setTabOrder(self.source_list, self.source_data_table)
        QWidget.setTabOrder(self.source_data_table, self.default_column_type_combo_box)
        QWidget.setTabOrder(self.default_column_type_combo_box, self.mapping_list)
        QWidget.setTabOrder(self.mapping_list, self.new_button)
        QWidget.setTabOrder(self.new_button, self.remove_button)
        QWidget.setTabOrder(self.remove_button, self.duplicate_button)
        QWidget.setTabOrder(self.duplicate_button, self.class_type_combo_box)
        QWidget.setTabOrder(self.class_type_combo_box, self.parameter_type_combo_box)
        QWidget.setTabOrder(self.parameter_type_combo_box, self.value_type_combo_box)
        QWidget.setTabOrder(self.value_type_combo_box, self.dimension_spin_box)
        QWidget.setTabOrder(self.dimension_spin_box, self.map_dimension_spin_box)
        QWidget.setTabOrder(self.map_dimension_spin_box, self.start_read_row_spin_box)
        QWidget.setTabOrder(self.start_read_row_spin_box, self.ignore_columns_button)
        QWidget.setTabOrder(self.ignore_columns_button, self.mapping_spec_table)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Import Editor", None))
        self.export_mappings_action.setText(QCoreApplication.translate("MainWindow", u"Export mappings...", None))
        self.import_mappings_action.setText(QCoreApplication.translate("MainWindow", u"Import mappings...", None))
        self.switch_input_type_action.setText(QCoreApplication.translate("MainWindow", u"Switch input type...", None))
#if QT_CONFIG(tooltip)
        self.switch_input_type_action.setToolTip(QCoreApplication.translate("MainWindow", u"Open a dialog to change input type.", None))
#endif // QT_CONFIG(tooltip)
        self.input_path_label.setText(QCoreApplication.translate("MainWindow", u"Input file/URL:", None))
        self.input_path_line_edit.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Select input file from the browse button...", None))
#if QT_CONFIG(tooltip)
        self.browse_inputs_button.setToolTip(QCoreApplication.translate("MainWindow", u"Browse input files or specify database URL.", None))
#endif // QT_CONFIG(tooltip)
        self.browse_inputs_button.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.input_type_label.setText(QCoreApplication.translate("MainWindow", u"Input type:", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Surplus column data type:", None))
#if QT_CONFIG(tooltip)
        self.default_column_type_combo_box.setToolTip(QCoreApplication.translate("MainWindow", u"Select data type for additional columns in variable-length pivoted source data.", None))
#endif // QT_CONFIG(tooltip)
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Mappings", None))
        self.new_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
        self.duplicate_button.setText(QCoreApplication.translate("MainWindow", u"Duplicate", None))
        self.class_type_combo_box.setItemText(0, QCoreApplication.translate("MainWindow", u"Entity class", None))
        self.class_type_combo_box.setItemText(1, QCoreApplication.translate("MainWindow", u"Entity group", None))
        self.class_type_combo_box.setItemText(2, QCoreApplication.translate("MainWindow", u"Alternative", None))
        self.class_type_combo_box.setItemText(3, QCoreApplication.translate("MainWindow", u"Scenario", None))
        self.class_type_combo_box.setItemText(4, QCoreApplication.translate("MainWindow", u"Scenario alternative", None))
        self.class_type_combo_box.setItemText(5, QCoreApplication.translate("MainWindow", u"Parameter value list", None))

        self.value_type_label.setText(QCoreApplication.translate("MainWindow", u"Default value type:", None))
        self.value_type_combo_box.setItemText(0, QCoreApplication.translate("MainWindow", u"Single value", None))
        self.value_type_combo_box.setItemText(1, QCoreApplication.translate("MainWindow", u"Time series", None))
        self.value_type_combo_box.setItemText(2, QCoreApplication.translate("MainWindow", u"Time pattern", None))
        self.value_type_combo_box.setItemText(3, QCoreApplication.translate("MainWindow", u"Map", None))
        self.value_type_combo_box.setItemText(4, QCoreApplication.translate("MainWindow", u"Array", None))

        self.class_type_label.setText(QCoreApplication.translate("MainWindow", u"Item type:", None))
#if QT_CONFIG(tooltip)
        self.dimension_spin_box.setToolTip(QCoreApplication.translate("MainWindow", u"Number of entity dimensions.", None))
#endif // QT_CONFIG(tooltip)
        self.parameter_type_label.setText(QCoreApplication.translate("MainWindow", u"Parameter type:", None))
        self.dimension_label.setText(QCoreApplication.translate("MainWindow", u"Number of dimensions:", None))
        self.ignore_columns_button.setText("")
        self.import_entity_alternatives_check_box.setText(QCoreApplication.translate("MainWindow", u"Import entity alternatives", None))
#if QT_CONFIG(tooltip)
        self.map_dimension_spin_box.setToolTip(QCoreApplication.translate("MainWindow", u"Number of dimensions when value type is Map.", None))
#endif // QT_CONFIG(tooltip)
        self.parameter_type_combo_box.setItemText(0, QCoreApplication.translate("MainWindow", u"Value", None))
        self.parameter_type_combo_box.setItemText(1, QCoreApplication.translate("MainWindow", u"Definition", None))
        self.parameter_type_combo_box.setItemText(2, QCoreApplication.translate("MainWindow", u"None", None))

        self.read_start_row_label.setText(QCoreApplication.translate("MainWindow", u"Read data from row:", None))
        self.ignore_columns_label.setText(QCoreApplication.translate("MainWindow", u"Ignore columns:", None))
        self.map_dimensions_label.setText(QCoreApplication.translate("MainWindow", u"Map dimensions:", None))
        self.import_entities_check_box.setText(QCoreApplication.translate("MainWindow", u"Import entities", None))
#if QT_CONFIG(tooltip)
        self.time_series_repeat_check_box.setToolTip(QCoreApplication.translate("MainWindow", u"Set the repeat flag for all imported time series", None))
#endif // QT_CONFIG(tooltip)
        self.time_series_repeat_check_box.setText(QCoreApplication.translate("MainWindow", u"Repeat time series", None))
        self.map_compression_check_box.setText(QCoreApplication.translate("MainWindow", u"Compress Maps", None))
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ImporterExecutable."""
import gc
from multiprocessing import Lock
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine.project_item.project_item_resource import database_resource, file_resource
from spine_items.importer.executable_item import ExecutableItem
from spine_items.importer.importer_specification import ImporterSpecification
from spinedb_api import DatabaseMapping, create_new_spine_database
from spinedb_api.spine_db_server import db_server_manager


class TestImporterExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        gc.collect()
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "Importer")

    def test_from_dict(self):
        item_dict = {
            "type": "Importer",
            "description": "",
            "x": 0,
            "y": 0,
            "specification": "importer_spec",
            "cancel_on_error": True,
            "on_conflict": "merge",
            "mapping_selection": [
                [{"type": "path", "relative": True, "path": ".spinetoolbox/items/data/units.xlsx"}, True]
            ],
        }
        spec_dict = {
            "name": "importer_spec",
            "description": "Some importer spec",
            "mapping": {
                "table_mappings": {
                    "Sheet1": [
                        {
                            "map_type": "ObjectClass",
                            "name": {"map_type": "column", "reference": 0},
                            "parameters": {"map_type": "None"},
                            "skip_columns": [],
                            "read_start_row": 0,
                            "objects": {"map_type": "column", "reference": 1},
                        }
                    ]
                },
                "table_options": {},
                "table_types": {"Sheet1": {"0": "string", "1": "string"}},
                "table_row_types": {},
                "selected_tables": ["Sheet1"],
                "source_type": "ExcelReader",
            },
        }
        logger = mock.MagicMock()
        specs = {"Importer": {"importer_spec": ImporterSpecification.from_dict(spec_dict, logger)}}
        item = ExecutableItem.from_dict(item_dict, "name", self._temp_dir.name, _MockSettings(), specs, logger)
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("Importer", item.item_type())

    def test_stop_execution(self):
        executable = ExecutableItem("name", {}, [], "", True, "merge", self._temp_dir.name, mock.MagicMock())
        executable.stop_execution()
        self.assertIsNone(executable._process)

    def test_execute_simplest_case(self):
        executable = ExecutableItem("name", {}, [], "", True, "merge", self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))
        # Check that _process is None after execution
        self.assertIsNone(executable._process)

    def test_execute_import_small_file(self):
        data_file = Path(self._temp_dir.name, "data.dat")
        self._write_simple_data(data_file)
        mapping = self._simple_input_data_mapping()
        database_path = Path(self._temp_dir.name, "database.sqlite")
        database_url = "sqlite:///" + str(database_path)
        create_new_spine_database(database_url)
        gams_path = ""
        logger = mock.MagicMock()
        logger.__reduce__ = lambda _: (mock.MagicMock, ())
        executable = ExecutableItem(
            "name", mapping, [str(data_file)], gams_path, True, "merge", self._temp_dir.name, logger
        )
        database_resources = [database_resource("provider", database_url)]
        file_resources = [file_resource("provider", str(data_file))]
        with db_server_manager() as mngr_queue:
            for r in database_resources:
                r.metadata["db_server_manager_queue"] = mngr_queue
            self.assertTrue(executable.execute(file_resources, database_resources, Lock()))
        # Check that _process is None after execution
        self.assertIsNone(executable._process)
        with DatabaseMapping(database_url) as database_map:
            class_list = database_map.query(database_map.entity_class_sq).all()
            self.assertEqual(len(class_list), 1)
            self.assertEqual(class_list[0].name, "class")
            entity_list = database_map.query(database_map.entity_sq).filter_by(class_id=class_list[0].id).all()
            self.assertEqual(len(entity_list), 1)
            self.assertEqual(entity_list[0].name, "entity")
        database_map.close()

    def test_execute_skip_deselected_file(self):
        data_file = Path(self._temp_dir.name, "data.dat")
        self._write_simple_data(data_file)
        database_path = Path(self._temp_dir.name, "database.sqlite")
        database_url = "sqlite:///" + str(database_path)
        create_new_spine_database(database_url)
        gams_path = ""
        executable = ExecutableItem("name", {}, [], gams_path, True, "merge", self._temp_dir.name, mock.MagicMock())
        database_resources = [database_resource("provider", database_url)]
        file_resources = [file_resource("provider", str(data_file))]
        self.assertTrue(executable.execute(file_resources, database_resources, Lock()))
        # Check that _process is None after execution
        self.assertIsNone(executable._process)
        with DatabaseMapping(database_url) as database_map:
            class_list = database_map.query(database_map.entity_class_sq).all()
            self.assertEqual(len(class_list), 0)
        database_map.close()

    @staticmethod
    def _write_simple_data(file_name):
        with open(file_name, "w") as out_file:
            out_file.write("class,entity\n")

    @staticmethod
    def _simple_input_data_mapping():
        return {
            "table_mappings": {
                "csv": [
                    {
                        "map_type": "ObjectClass",
                        "name": {"map_type": "column", "reference": 0},
                        "parameters": {"map_type": "None"},
                        "skip_columns": [],
                        "read_start_row": 0,
                        "objects": {"map_type": "column", "reference": 1},
                    }
                ]
            },
            "table_options": {
                "csv": {
                    "encoding": "ascii",
                    "delimeter": ",",
                    "delimiter_custom": "",
                    "quotechar": '"',
                    "skip_header": False,
                    "skip": 0,
                }
            },
            "table_types": {"csv": {0: "string", 1: "string"}},
            "table_row_types": {},
            "selected_tables": ["csv"],
            "source_type": "CSVReader",
        }


class _MockSettings:
    @staticmethod
    def value(key, defaultValue=None):
        return {"appSettings/gamsPath": ""}.get(key, defaultValue)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""ImportMappingOptions widget."""
from PySide6.QtCore import QModelIndex, Qt, Slot
from ..commands import (
    SetImportEntitiesFlag,
    SetImportEntityAlternatives,
    SetItemMappingDimensionCount,
    SetItemMappingType,
    SetMapCompressFlag,
    SetMapDimensionCount,
    SetParameterType,
    SetReadStartRow,
    SetSkipColumns,
    SetTimeSeriesRepeatFlag,
    SetValueType,
)
from ..flattened_mappings import MappingType
from ..mvcmodels.mappings_model import Role
from .custom_menus import SimpleFilterMenu


class ImportMappingOptions:
    """
    Controls the 'Mapping options' part of the window.
    """

    def __init__(self, mappings_model, ui, undo_stack):
        """
        Args:
            mappings_model (MappingsModel): mappings model
            ui (Any): import editor's UI
            undo_stack (QUndoStack): undo stack
        """
        # state
        self._mappings_model = mappings_model
        self._list_index = QModelIndex()
        self._ui = ui
        self._undo_stack = undo_stack
        self._block_signals = False
        ## ui
        self._ui_ignore_columns_filtermenu = SimpleFilterMenu(self._ui.ignore_columns_button, show_empty=False)
        self._ui.ignore_columns_button.setMenu(self._ui_ignore_columns_filtermenu)
        ## connect signals
        self._mappings_model.dataChanged.connect(self._update_options)
        self._ui.mapping_list.selectionModel().currentChanged.connect(self.reload_options)
        self._ui.dimension_spin_box.valueChanged.connect(self._change_dimension_count)
        self._ui.class_type_combo_box.currentTextChanged.connect(self._change_item_mapping_type)
        self._ui.parameter_type_combo_box.currentTextChanged.connect(self._change_parameter_type)
        self._ui.value_type_combo_box.currentTextChanged.connect(self._change_value_type)
        self._ui.import_entities_check_box.stateChanged.connect(self._change_import_entities)
        self._ui_ignore_columns_filtermenu.filterChanged.connect(self._change_skip_columns)
        self._ui.start_read_row_spin_box.valueChanged.connect(self._change_read_start_row)
        self._ui.import_entity_alternatives_check_box.stateChanged.connect(self._change_import_entity_alternatives)
        self._ui.time_series_repeat_check_box.stateChanged.connect(self._change_time_series_repeat_flag)
        self._ui.map_dimension_spin_box.valueChanged.connect(self._change_map_dimension_count)
        self._ui.map_compression_check_box.stateChanged.connect(self._change_map_compression_flag)
        self._update_ui()

    @Slot(int)
    def set_num_available_columns(self, num):
        mapping_index = self._ui.mapping_list.selectionModel().currentIndex()
        if mapping_index.isValid():
            selected = mapping_index.data(Role.ITEM).flattened_mappings.root_mapping.skip_columns
        else:
            selected = []
        # The filter menu is 1-based
        self._ui_ignore_columns_filtermenu._filter._filter_model.set_list(set(range(1, num + 1)))
        self._update_ignore_columns_button(selected)

    def _has_current_mappings(self):
        """Checks if mappings list has current mappings.

        Returns:
            bool: True if mappings list has current selected, False otherwise
        """
        if not self._list_index.isValid():
            return False
        table_index = self._ui.source_list.selectionModel().currentIndex()
        if not table_index.isValid() or table_index.row() < 1:
            return False
        return self._mappings_model.rowCount(table_index) > 0

    @Slot(QModelIndex, QModelIndex, int)
    def _update_options(self, top_left, bottom_right, roles):
        """Updates widget contents when flattened mappings change.

        Args:
            top_left (QModelIndex): top left index
            bottom_right (QModelIndex): bottom right index
            roles (list of int): Qt's data roles
        """
        if bottom_right.column() > 0 and top_left.parent() == self._list_index:
            flattened_mappings = self._list_index.data(Role.FLATTENED_MAPPINGS)
            self._ui.ignore_columns_button.setEnabled(flattened_mappings.root_mapping.is_pivoted())
            self._ui.ignore_columns_label.setEnabled(flattened_mappings.root_mapping.is_pivoted())
            return
        if top_left != self._list_index or Role.FLATTENED_MAPPINGS not in roles:
            return
        self._update_ui()

    @Slot(QModelIndex, QModelIndex)
    def reload_options(self, current, previous):
        """Reloads widget contents.

        Args:
            current (QModelIndex): currently selected mapping list index
            previous (QModelIndex): previously selected mapping list index
        """
        self._list_index = current
        self._update_ui()

    def _update_ui(self):
        """
        Updates ui according to the current mapping type.
        """
        has_mapping = self._has_current_mappings()
        self._ui.mapping_options_contents.setEnabled(has_mapping)
        for i in range(self._ui.mapping_options_contents.count()):
            self._ui.mapping_options_contents.itemAt(i).widget().setEnabled(has_mapping)
        if not has_mapping:
            return
        flattened_mappings = self._list_index.data(Role.FLATTENED_MAPPINGS)
        self._block_signals = True
        try:
            class_type_index = [
                MappingType.EntityClass,
                MappingType.EntityGroup,
                MappingType.Alternative,
                MappingType.Scenario,
                MappingType.ScenarioAlternative,
                MappingType.ParameterValueList,
            ].index(flattened_mappings.map_type)
        except ValueError:
            class_type_index = -1
        self._ui.class_type_combo_box.setCurrentIndex(class_type_index)

        # update item mapping settings
        if flattened_mappings.may_import_entity_alternatives():
            self._ui.import_entity_alternatives_check_box.setEnabled(True)
            check_state = (
                Qt.CheckState.Checked if flattened_mappings.import_entity_alternatives() else Qt.CheckState.Unchecked
            )
            self._ui.import_entity_alternatives_check_box.setCheckState(check_state)
        else:
            self._ui.import_entity_alternatives_check_box.setEnabled(False)
            self._ui.import_entity_alternatives_check_box.setCheckState(Qt.CheckState.Unchecked)
        if flattened_mappings.may_import_entities():
            self._ui.import_entities_check_box.setEnabled(True)
            check_state = Qt.CheckState.Checked if flattened_mappings.import_entities() else Qt.CheckState.Unchecked
            self._ui.import_entities_check_box.setCheckState(check_state)
        else:
            self._ui.import_entities_check_box.setEnabled(False)
            self._ui.import_entities_check_box.setCheckState(Qt.CheckState.Unchecked)
        can_have_dimensions = flattened_mappings.can_have_dimensions()
        self._ui.dimension_label.setEnabled(can_have_dimensions)
        self._ui.dimension_spin_box.setEnabled(can_have_dimensions)
        if can_have_dimensions:
            self._ui.dimension_spin_box.setValue(flattened_mappings.dimension_count())

        # update parameter mapping settings
        has_parameters = flattened_mappings.has_parameters()
        self._ui.parameter_type_label.setEnabled(has_parameters)
        self._ui.parameter_type_combo_box.setEnabled(has_parameters)
        if has_parameters:
            self._ui.parameter_type_combo_box.setCurrentText(flattened_mappings.display_parameter_type())
        has_value_component = flattened_mappings.has_value_component()
        self._ui.value_type_label.setEnabled(has_value_component)
        self._ui.value_type_combo_box.setEnabled(has_value_component)
        if has_value_component:
            self._ui.value_type_combo_box.setCurrentText(flattened_mappings.value_type)
            self._ui.value_type_label.setText(flattened_mappings.value_type_label())

        # update ignore columns filter
        self._ui.ignore_columns_button.setEnabled(flattened_mappings.root_mapping.is_pivoted())
        self._ui.ignore_columns_label.setEnabled(flattened_mappings.root_mapping.is_pivoted())
        self._update_ignore_columns_button(flattened_mappings.root_mapping.skip_columns)

        self._ui.start_read_row_spin_box.setValue(flattened_mappings.root_mapping.read_start_row + 1)

        self._update_time_series_options()
        self._update_map_options()
        self._block_signals = False

    @Slot(list)
    def _update_ignore_columns_button(self, skip_cols):
        """
        Args:
            skip_cols (list of int): 0-based list of ignored columns
        """
        # NOTE: We go from 0-based to 1-based, for visualization
        skip_cols = [c + 1 for c in skip_cols]
        self._ui_ignore_columns_filtermenu._filter._filter_model.set_selected(skip_cols)
        skip_button_text = ", ".join(str(c) for c in skip_cols)
        if len(skip_button_text) > 20:
            skip_button_text = skip_button_text[:20] + "..."
        self._ui.ignore_columns_button.setText(skip_button_text)

    @Slot(str)
    def _change_item_mapping_type(self, new_type):
        """
        Pushes a SetItemMappingType command to the undo stack

        Args:
            new_type (str): item's new type
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_mapping = self._list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._undo_stack.push(
            SetItemMappingType(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                new_type,
                previous_mapping,
            )
        )

    @Slot(int)
    def _change_dimension_count(self, dimension_count):
        """
        Pushes a SetItemMappingDimensionCount command to the undo stack.

        Args:
            dimension_count (int): mapping's dimension
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_mapping = self._list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._undo_stack.push(
            SetItemMappingDimensionCount(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                dimension_count,
                previous_mapping,
            )
        )

    @Slot(str)
    def _change_parameter_type(self, new_type):
        """
        Pushes a SetParameterType command to undo stack.

        Args:
            new_type (str): new parameter type's name
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_mapping = self._list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._undo_stack.push(
            SetParameterType(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                new_type,
                previous_mapping,
            )
        )

    @Slot(str)
    def _change_value_type(self, new_type):
        """
        Pushes a SetValueType command to undo stack.

        Args:
            new_type (str): new value type's name
        """
        if self._block_signals or not self._has_current_mappings():
            return
        old_type = self._list_index.data(Role.ITEM).flattened_mappings.value_type
        self._undo_stack.push(
            SetValueType(
                self._list_index.parent().row(), self._list_index.row(), self._mappings_model, new_type, old_type
            )
        )

    @Slot(int)
    def _change_import_entities(self, state):
        """
        Pushes SetImportEntitiesFlag command to the undo stack.

        Args:
            state (int): New state value
        """
        if self._block_signals or not self._has_current_mappings():
            return
        self._undo_stack.push(
            SetImportEntitiesFlag(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                state == Qt.CheckState.Checked.value,
            )
        )

    @Slot(int)
    def _change_read_start_row(self, row):
        """
        Pushes :class:`SetReadStartRow` to the undo stack.

        Args:
            row (int): new read start row
        """
        if self._block_signals or not self._has_current_mappings():
            return
        row -= 1
        previous_row = self._list_index.data(Role.ITEM).flattened_mappings.read_start_row()
        self._undo_stack.push(
            SetReadStartRow(
                self._list_index.parent().row(), self._list_index.row(), self._mappings_model, row, previous_row
            )
        )

    @Slot(int)
    def _change_import_entity_alternatives(self, state):
        """
        Pushes SetImportEntityAlternatives command to the undo stack.

        Args:
            state (Qt.CheckState): New check box state
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_mapping_root = self._list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._undo_stack.push(
            SetImportEntityAlternatives(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                state == Qt.CheckState.Checked.value,
                previous_mapping_root,
            )
        )

    def _change_skip_columns(self, skip_cols):
        """Pushes :class:`SetSkipColumns` to the undo stack.

        Args:
            skip_cols (list): list of columns or column names
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_skip_cols = self._list_index.data(Role.ITEM).flattened_mappings.skip_columns().copy()
        # NOTE: The columns in the filter menu are 1-based, for visualization. Here we need them 0-based
        skip_cols = [c - 1 for c in skip_cols]
        self._undo_stack.push(
            SetSkipColumns(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                skip_cols,
                previous_skip_cols,
            )
        )

    @Slot(int)
    def _change_time_series_repeat_flag(self, repeat):
        """
        Pushes :class:`SetTimeSeriesRepeatFlag` to the undo stack.

        Args:
            repeat (int): True if repeat is enabled, False otherwise
        """
        if self._block_signals or not self._has_current_mappings():
            return
        self._undo_stack.push(
            SetTimeSeriesRepeatFlag(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                repeat == Qt.CheckState.Checked.value,
            )
        )

    @Slot(int)
    def _change_map_dimension_count(self, dimension_count):
        """
        Pushes :class:`SetMapDimensionCount` to the undo stack.

        Args:
            dimension_count (int): new map dimension_count
        """
        if self._block_signals or not self._has_current_mappings():
            return
        previous_mapping_root = self._list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._undo_stack.push(
            SetMapDimensionCount(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                dimension_count,
                previous_mapping_root,
            )
        )

    @Slot(int)
    def _change_map_compression_flag(self, compress):
        """
        Pushes :class:`SetMapCompressFlag` to the undo stack.

        Args:
            compress (int): if ``Qt.CheckState.Checked.value``, Maps will be compressed
        """
        if self._block_signals or not self._has_current_mappings():
            return
        self._undo_stack.push(
            SetMapCompressFlag(
                self._list_index.parent().row(),
                self._list_index.row(),
                self._mappings_model,
                compress == Qt.CheckState.Checked.value,
            )
        )

    def _update_time_series_options(self):
        """Updates widgets that concern time series type parameters"""
        if not self._has_current_mappings():
            return
        flattened_mappings = self._list_index.data(Role.ITEM).flattened_mappings
        value_mapping = flattened_mappings.value_mapping()
        if value_mapping is None:
            self._ui.time_series_repeat_check_box.setEnabled(False)
            return
        is_time_series = flattened_mappings.is_time_series_value()
        self._ui.time_series_repeat_check_box.setEnabled(is_time_series)
        self._ui.time_series_repeat_check_box.setCheckState(
            Qt.CheckState.Checked if is_time_series and value_mapping.options.get("repeat") else Qt.CheckState.Unchecked
        )

    def _update_map_options(self):
        """Updates widgets that concern map type parameters."""
        if not self._has_current_mappings():
            return
        flattened_mappings = self._list_index.data(Role.ITEM).flattened_mappings
        value_mapping = flattened_mappings.value_mapping()
        if value_mapping is None:
            self._ui.map_dimensions_label.setEnabled(False)
            self._ui.map_dimension_spin_box.setEnabled(False)
            self._ui.map_compression_check_box.setEnabled(False)
            return
        is_map = flattened_mappings.is_map_value()
        dimension_count = flattened_mappings.map_dimension_count()
        self._ui.map_dimensions_label.setEnabled(is_map)
        self._ui.map_dimension_spin_box.setEnabled(is_map)
        self._ui.map_dimension_spin_box.setValue(dimension_count)
        self._ui.map_compression_check_box.setEnabled(is_map)
        self._ui.map_compression_check_box.setChecked(bool(is_map and value_mapping.compress))



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Undo/redo commands for the DataConnection project item."""
from pathlib import Path
from spine_items.commands import SpineToolboxCommand


class AddDCReferencesCommand(SpineToolboxCommand):
    """Command to add DC references."""

    def __init__(self, dc_name, file_refs, db_refs, project):
        """
        Args:
            dc_name (str): DC name
            file_refs (list of str): list of file refs to add
            db_refs (list of str): list of db refs to add
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._dc_name = dc_name
        self._file_refs = file_refs
        self._db_refs = db_refs
        self._project = project
        self.setText(f"add references to {dc_name}")

    def redo(self):
        dc = self._project.get_item(self._dc_name)
        dc.do_add_references(self._file_refs, self._db_refs)

    def undo(self):
        dc = self._project.get_item(self._dc_name)
        dc.do_remove_references(self._file_refs, self._db_refs)


class RemoveDCReferencesCommand(SpineToolboxCommand):
    """Command to remove DC references."""

    def __init__(self, dc_name, file_refs, db_refs, project):
        """
        Args:
            dc_name (str): DC name
            file_refs (list of str): list of file refs to remove
            db_refs (list of str): list of db refs to remove
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._dc_name = dc_name
        self._file_refs = file_refs
        self._db_refs = db_refs
        self._project = project
        self.setText(f"remove references from {dc_name}")

    def redo(self):
        dc = self._project.get_item(self._dc_name)
        dc.do_remove_references(self._file_refs, self._db_refs)

    def undo(self):
        dc = self._project.get_item(self._dc_name)
        dc.do_add_references(self._file_refs, self._db_refs)


class MoveReferenceToData(SpineToolboxCommand):
    """Command to move DC references to data."""

    def __init__(self, dc_name, paths, project):
        """
        Args:
            dc_name (str): DC name
            paths (list of str): list of paths to move
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._dc_name = dc_name
        self._paths = paths
        self._project = project
        self.setText(f"copy references to data in {dc_name}")

    def redo(self):
        dc = self._project.get_item(self._dc_name)
        dc.do_copy_to_project(self._paths)
        dc.do_remove_references(self._paths, [])

    def undo(self):
        dc = self._project.get_item(self._dc_name)
        dc.delete_files_from_project([Path(p).name for p in self._paths])
        dc.do_add_references(self._paths, [])



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``models`` module."""
from pathlib import Path
import unittest
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication
from spine_engine.project_item.project_item_resource import file_resource, file_resource_in_pack
from spine_items.models import CheckableFileListModel


class TestCheckableFileListModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._model = CheckableFileListModel()

    def tearDown(self):
        self._model.deleteLater()

    def test_update(self):
        single_resource = file_resource("item name", str(Path.cwd() / "path" / "to" / "file"), "file label")
        pack_resource = file_resource_in_pack("item name", "pack label", str(Path.cwd() / "path" / "to" / "pack_file"))
        self._model.update([single_resource, pack_resource])
        self.assertEqual(self._model.rowCount(), 2)
        index = self._model.index(0, 0)
        self.assertEqual(index.data(), "file label")
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), str(Path.cwd() / "path" / "to" / "file"))
        self.assertEqual(self._model.rowCount(index), 0)
        index = self._model.index(1, 0)
        self.assertEqual(index.data(), "pack label")
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), None)
        self.assertEqual(self._model.rowCount(index), 1)
        index = self._model.index(0, 0, index)
        self.assertEqual(index.data(), str(Path.cwd() / "path" / "to" / "pack_file"))
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), str(Path.cwd() / "path" / "to" / "pack_file"))

    def test_replace_single_resource(self):
        single_resource = file_resource("item name", str(Path.cwd() / "path" / "to" / "file"), "file label")
        self._model.update([single_resource])
        new_resource = file_resource("other item", str(Path.cwd() / "path" / "to" / "another_file"), "new label")
        self._model.replace(single_resource, new_resource)
        self.assertEqual(self._model.rowCount(), 1)
        index = self._model.index(0, 0)
        self.assertEqual(index.data(), "new label")
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), str(Path.cwd() / "path" / "to" / "another_file"))
        self.assertEqual(self._model.rowCount(index), 0)

    def test_replace_pack_resource_file_path(self):
        pack_resource = file_resource_in_pack("item name", "pack label", str(Path.cwd() / "path" / "to" / "pack_file"))
        self._model.update([pack_resource])
        new_resource = file_resource_in_pack("item name", "pack label", str(Path.cwd() / "path" / "to" / "pack_file_2"))
        self._model.replace(pack_resource, new_resource)
        self.assertEqual(self._model.rowCount(), 1)
        index = self._model.index(0, 0)
        self.assertEqual(index.data(), "pack label")
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), None)
        self.assertEqual(self._model.rowCount(index), 1)
        index = self._model.index(0, 0, index)
        self.assertEqual(index.data(), str(Path.cwd() / "path" / "to" / "pack_file_2"))
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), str(Path.cwd() / "path" / "to" / "pack_file_2"))

    def test_replace_pack_resource_with_completely_new_one(self):
        pack_resource = file_resource_in_pack("item name", "pack label", str(Path.cwd() / "path" / "to" / "pack_file"))
        self._model.update([pack_resource])
        new_resource = file_resource_in_pack(
            "another item", "new pack label", str(Path.cwd() / "path" / "to" / "pack_file_2")
        )
        self._model.replace(pack_resource, new_resource)
        self.assertEqual(self._model.rowCount(), 1)
        index = self._model.index(0, 0)
        self.assertEqual(index.data(), "new pack label")
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), None)
        self.assertEqual(self._model.rowCount(index), 1)
        index = self._model.index(0, 0, index)
        self.assertEqual(index.data(), str(Path.cwd() / "path" / "to" / "pack_file_2"))
        self.assertEqual(index.data(Qt.ItemDataRole.ToolTipRole), str(Path.cwd() / "path" / "to" / "pack_file_2"))


if __name__ == "__main__":
    unittest.main()
